import pytest
import yaml

from substra.sdk import exceptions
from substra.sdk import schemas
from substra.sdk.client import Client
from substra.sdk.client import _upper_slug
from substra.sdk.exceptions import ConfigurationInfoError


@pytest.fixture
def config_file(tmp_path):
    config_dict = {
        "toto": {
            "backend_type": "remote",
            "url": "toto-org.com",
            "username": "toto_file_username",
            "password": "toto_file_password",
        }
    }
    config_file = tmp_path / "config.yaml"
    config_file.write_text(yaml.dump(config_dict))
    return config_file


def stub_login(username, password):
    if username == "org-1" and password == "password1":
        return "token1"
    if username == "env_var_username" and password == "env_var_password":
        return "env_var_token"
    if username == "toto_file_username" and password == "toto_file_password":
        return "toto_file_token"


@pytest.mark.parametrize(
    ["input", "expected"],
    [
        ("toto", "TOTO"),
        ("client-org-1", "CLIENT_ORG_1"),
        ("un nom très français", "UN_NOM_TRES_FRANCAIS"),
    ],
)
def test_upper_slug(input, expected):
    assert expected == _upper_slug(input)


def test_default_client():
    client = Client()
    assert client.backend_mode == schemas.BackendType.LOCAL_SUBPROCESS


@pytest.mark.parametrize(
    ["mode", "client_name", "url", "token", "insecure", "retry_timeout"],
    [
        (schemas.BackendType.LOCAL_SUBPROCESS, "foo", None, None, True, 5),
        (schemas.BackendType.LOCAL_DOCKER, "foobar", None, None, True, None),
        (schemas.BackendType.REMOTE, "bar", "example.com", "bloop", False, 15),
        (schemas.BackendType.LOCAL_SUBPROCESS, "hybrid", "example.com", "foo", True, 500),
        (schemas.BackendType.REMOTE, "foofoo", "https://example.com/api-token-auth/", None, True, 500),
    ],
)
def test_client_configured_in_code(mode, client_name, url, token, insecure, retry_timeout):
    client = Client(
        backend_type=mode,
        client_name=client_name,
        url=url,
        token=token,
        insecure=insecure,
        retry_timeout=retry_timeout,
    )
    assert client.backend_mode == mode
    assert client._insecure == insecure
    if retry_timeout is not None:
        assert client._retry_timeout == retry_timeout
    else:
        assert client._retry_timeout == 300
    if token is not None:
        assert client._token == token
    else:
        assert client._token is None
    if url is not None:
        assert client._url == url
    else:
        assert client._url is None


def test_client_should_raise_when_missing_name():
    with pytest.raises(ConfigurationInfoError):
        Client(configuration_file="something")


def test_client_with_password(mocker):
    mocker.patch("substra.sdk.Client.login", side_effect=stub_login)
    rest_client_logout = mocker.patch("substra.sdk.backends.remote.rest_client.Client.logout")
    client_args = {
        "backend_type": "remote",
        "url": "example.com",
        "token": None,
        "username": "org-1",
        "password": "password1",
    }

    client = Client(**client_args)
    assert client._token == "token1"
    client.logout()
    assert client._token is None
    rest_client_logout.assert_called_once()
    del client

    rest_client_logout.reset_mock()
    client = Client(**client_args)
    del client
    rest_client_logout.assert_called_once()

    rest_client_logout.reset_mock()
    with Client(**client_args) as client:
        assert client._token == "token1"
    rest_client_logout.assert_called_once()


def test_client_token_supercedes_password(mocker):
    mocker.patch("substra.sdk.Client.login", side_effect=stub_login)
    client = Client(
        backend_type="remote",
        url="example.com",
        token="token0",
        username="org-1",
        password="password1",
    )
    assert client._token == "token0"


def test_client_configuration_from_env_var(mocker, monkeypatch):
    mocker.patch("substra.sdk.Client.login", side_effect=stub_login)
    monkeypatch.setenv("SUBSTRA_TOTO_BACKEND_TYPE", "remote")
    monkeypatch.setenv("SUBSTRA_TOTO_URL", "toto-org.com")
    monkeypatch.setenv("SUBSTRA_TOTO_USERNAME", "env_var_username")
    monkeypatch.setenv("SUBSTRA_TOTO_PASSWORD", "env_var_password")
    monkeypatch.setenv("SUBSTRA_TOTO_RETRY_TIMEOUT", "42")
    monkeypatch.setenv("SUBSTRA_TOTO_INSECURE", "true")
    client = Client(client_name="toto")
    assert client.backend_mode == "remote"
    assert client._url == "toto-org.com"
    assert client._token == "env_var_token"
    assert client._retry_timeout == 42
    assert client._insecure is True


def test_client_configuration_from_config_file(mocker, config_file):
    mocker.patch("substra.sdk.Client.login", side_effect=stub_login)
    client = Client(configuration_file=config_file, client_name="toto")
    assert client.backend_mode == "remote"
    assert client._url == "toto-org.com"
    assert client._token == "toto_file_token"
    assert client._retry_timeout == 300
    assert client._insecure is False


def test_client_configuration_code_overrides_env_var(monkeypatch):
    """
    A variable set in the code overrides one set in an env variable
    """
    monkeypatch.setenv("SUBSTRA_TOTO_BACKEND_TYPE", "remote")
    monkeypatch.setenv("SUBSTRA_TOTO_URL", "toto-org.com")
    monkeypatch.setenv("SUBSTRA_TOTO_USERNAME", "env_var_username")
    monkeypatch.setenv("SUBSTRA_TOTO_PASSWORD", "env_var_password")
    monkeypatch.setenv("SUBSTRA_TOTO_RETRY_TIMEOUT", "42")
    monkeypatch.setenv("SUBSTRA_TOTO_INSECURE", "true")
    client = Client(
        client_name="toto",
        backend_type="subprocess",
        url="",
    )
    assert client.backend_mode == "subprocess"
    assert client._url == ""
    assert client._token is None
    assert client._retry_timeout == 42
    assert client._insecure is True


def test_client_configuration_code_overrides_config_file(mocker, config_file):
    """
    A variable set in the code overrides one set in a config file
    """
    mocker.patch("substra.sdk.Client.login", side_effect=stub_login)
    client = Client(
        client_name="toto",
        configuration_file=config_file,
        username="org-1",
        password="password1",
        retry_timeout=100,
    )
    assert client.backend_mode == "remote"
    assert client._url == "toto-org.com"
    assert client._token == "token1"
    assert client._retry_timeout == 100
    assert client._insecure is False


def test_client_configuration_env_var_overrides_config_file(mocker, monkeypatch, config_file):
    """
    A variable set in an env var overrides one set in a config file
    """
    mocker.patch("substra.sdk.Client.login", side_effect=stub_login)
    monkeypatch.setenv("SUBSTRA_TOTO_BACKEND_TYPE", "docker")
    monkeypatch.setenv("SUBSTRA_TOTO_USERNAME", "env_var_username")
    monkeypatch.setenv("SUBSTRA_TOTO_PASSWORD", "env_var_password")
    client = Client(configuration_file=config_file, client_name="toto", retry_timeout=12)
    assert client.backend_mode == "docker"
    assert client._url == "toto-org.com"
    assert client._token == "env_var_token"
    assert client._retry_timeout == 12
    assert client._insecure is False


def test_login_remote_without_url(tmpdir):
    with pytest.raises(exceptions.SDKException):
        Client(backend_type="remote")


def test_client_configuration_configuration_file_path_from_env_var(mocker, monkeypatch, config_file):
    """
    The configuration file path can be set through an env var
    """
    mocker.patch("substra.sdk.Client.login", side_effect=stub_login)
    monkeypatch.setenv("SUBSTRA_CLIENTS_CONFIGURATION_FILE_PATH", config_file)
    client = Client(client_name="toto")
    assert client.backend_mode == "remote"
    assert client._url == "toto-org.com"
    assert client._token == "toto_file_token"
    assert client._retry_timeout == 300
    assert client._insecure is False


def test_client_configuration_configuration_file_path_parameter_supercedes_env_var(
    mocker, monkeypatch, config_file, tmp_path
):
    """
    The configuration file path env var is supercedes by `configuration_file=`
    """
    mocker.patch("substra.sdk.Client.login", side_effect=stub_login)
    monkeypatch.setenv("SUBSTRA_CLIENTS_CONFIGURATION_FILE_PATH", config_file)

    config_2_dict = {
        "toto": {
            "backend_type": "docker",
        }
    }
    config_2_file = tmp_path / "config.yaml"
    config_2_file.write_text(yaml.dump(config_2_dict))

    client = Client(configuration_file=config_2_file, client_name="toto")
    assert client.backend_mode == "docker"


def test_client_configuration_file_path_env_var_empty_string(mocker, monkeypatch, config_file):
    """
    The configuration file path env var is supercedes by `configuration_file=`
    """
    mocker.patch("substra.sdk.Client.login", side_effect=stub_login)
    monkeypatch.setenv("SUBSTRA_CLIENTS_CONFIGURATION_FILE_PATH", config_file)

    client = Client(configuration_file="", client_name="toto")
    assert client.backend_mode == "subprocess"



import logging
from typing import Union

import requests

logger = logging.getLogger(__name__)


class SDKException(Exception):
    pass


class LoadDataException(SDKException):
    pass


class RequestException(SDKException):
    def __init__(self, msg, status_code):
        self.msg = msg
        self.status_code = status_code
        super().__init__(msg)

    @classmethod
    def from_request_exception(cls, request_exception):
        msg = None
        try:
            msg = request_exception.response.json()["detail"]
            msg = f"{request_exception}: {msg}"
        except Exception:
            msg = str(request_exception)

        try:
            status_code = request_exception.response.status_code
        except AttributeError:
            status_code = None

        return cls(msg, status_code)


class ConnectionError(RequestException):
    pass


class Timeout(RequestException):
    pass


class HTTPError(RequestException):
    pass


class InternalServerError(HTTPError):
    pass


class GatewayUnavailable(HTTPError):
    pass


class InvalidRequest(HTTPError):
    def __init__(self, msg, status_code, errors=None):
        super().__init__(msg, status_code)
        self.errors = errors

    @classmethod
    def from_request_exception(cls, request_exception):
        try:
            error = request_exception.response.json()
        except ValueError:
            error = request_exception.response

        get_method = getattr(error, "get", None)
        if callable(get_method):
            msg = get_method("detail", str(error))
        else:
            msg = str(error)

        try:
            status_code = request_exception.response.status_code
        except AttributeError:
            status_code = None

        return cls(msg, status_code, error)


class NotFound(HTTPError):
    pass


class RequestTimeout(HTTPError):
    def __init__(self, key, status_code):
        self.key = key
        msg = f"Operation on object with key(s) '{key}' timed out."
        super().__init__(msg, status_code)

    @classmethod
    def from_request_exception(cls, request_exception):
        # parse response and fetch key
        r = request_exception.response.json()

        try:
            key = r["key"] if "key" in r else r["detail"].get("key")
        except (AttributeError, KeyError):
            # XXX this is the case when doing a POST query to update the
            #     data manager for instance
            key = None

        return cls(key, request_exception.response.status_code)


class AlreadyExists(HTTPError):
    def __init__(self, key, status_code):
        self.key = key
        msg = f"Object with key(s) '{key}' already exists."
        super().__init__(msg, status_code)

    @classmethod
    def from_request_exception(cls, request_exception):
        # parse response and fetch key
        r = request_exception.response.json()
        # XXX support list of keys; this could be the case when adding
        #     a list of data samples through a single POST request
        if isinstance(r, list):
            key = [x["key"] for x in r]
        elif isinstance(r, dict):
            key = r.get("key", None)
        else:
            key = r

        return cls(key, request_exception.response.status_code)


class InvalidResponse(SDKException):
    def __init__(self, response, msg):
        self.response = response
        super(InvalidResponse, self).__init__(msg)


class AuthenticationError(HTTPError):
    pass


class AuthorizationError(HTTPError):
    pass


class BadLoginException(RequestException):
    """The server refused to log-in with these credentials"""

    pass


class UsernamePasswordLoginDisabledException(RequestException):
    """The server disabled the endpoint, preventing the use of Client.login"""

    @classmethod
    def from_request_exception(cls, request_exception):
        base = super().from_request_exception(request_exception)
        return cls(
            base.msg
            + (
                "\n\nAuthenticating with username/password is disabled.\n"
                "Log onto the frontend for your instance and generate a token there, "
                'then use it in the Client(token="...") constructor: '
                "https://docs.substra.org/en/stable/documentation/api_tokens_generation.html"
            ),
            base.status_code,
        )


def from_request_exception(
    e: requests.exceptions.RequestException,
) -> Union[RequestException, requests.exceptions.RequestException]:
    """
    try turning an exception from the `requests` library into a Substra exception
    """
    connection_error_mapping: dict[requests.exceptions.RequestException, RequestException] = {
        requests.exceptions.ConnectionError: ConnectionError,
        requests.exceptions.Timeout: Timeout,
    }
    for k, v in connection_error_mapping.items():
        if isinstance(e, k):
            return v.from_request_exception(e)

    http_status_mapping: dict[int, RequestException] = {
        400: InvalidRequest,
        401: AuthenticationError,
        403: AuthorizationError,
        404: NotFound,
        408: RequestTimeout,
        409: AlreadyExists,
        500: InternalServerError,
        502: GatewayUnavailable,
        503: GatewayUnavailable,
        504: GatewayUnavailable,
    }
    if isinstance(e, requests.exceptions.HTTPError):
        logger.error(f"Requests error status {e.response.status_code}: {e.response.text}")
        return http_status_mapping.get(e.response.status_code, HTTPError).from_request_exception(e)

    return e


class ConfigurationInfoError(SDKException):
    """ConfigurationInfoError"""

    pass


class BadConfiguration(SDKException):
    """Bad configuration"""

    pass


class UserException(SDKException):
    """User Exception"""

    pass


class EmptyInModelException(SDKException):
    """No in_models when needed"""

    pass


class ComputePlanKeyFormatError(Exception):
    """The given compute plan key has to respect the UUID format."""

    pass


class OrderingFormatError(Exception):
    """The given ordering parameter has to respect expected format."""

    pass


class FilterFormatError(Exception):
    """The given filters has to respect expected format."""

    pass


class NotAllowedFilterError(Exception):
    """The given filter is not available on asset."""

    pass


class KeyAlreadyExistsError(Exception):
    """The asset key has already been used."""

    pass


class _TaskAssetError(Exception):
    """Base exception class for task asset error"""

    def __init__(self, *, compute_task_key: str, identifier: str, message: str):
        self.compute_task_key = compute_task_key
        self.identifier = identifier
        self.message = message
        super().__init__(self.message)

    pass


class TaskAssetNotFoundError(_TaskAssetError):
    """Exception raised when no task input/output asset have not been found for specific task key and identifier"""

    def __init__(self, compute_task_key: str, identifier: str):
        message = f"No task asset found with {compute_task_key=} and {identifier=}"
        super().__init__(compute_task_key=compute_task_key, identifier=identifier, message=message)

    pass


class TaskAssetMultipleFoundError(_TaskAssetError):
    """Exception raised when more than one task input/output assets have been found for specific task key and
    identifier"""

    def __init__(self, compute_task_key: str, identifier: str):
        message = f"Multiple task assets found with {compute_task_key=} and {identifier=}"
        super().__init__(compute_task_key=compute_task_key, identifier=identifier, message=message)


class FutureError(Exception):
    """Error while waiting a blocking operation to complete"""


class FutureTimeoutError(FutureError):
    """Future execution timed out."""


class FutureFailureError(FutureError):
    """Future execution failed."""



import pytest

import substra


@pytest.fixture(scope="session")
def docker_clients():
    return [substra.Client(backend_type=substra.BackendType.LOCAL_DOCKER) for _ in range(2)]


@pytest.fixture(scope="session")
def subprocess_clients():
    return [substra.Client(backend_type=substra.BackendType.LOCAL_SUBPROCESS) for _ in range(2)]


@pytest.fixture(scope="session", params=["docker", "subprocess"])
def clients(request, docker_clients, subprocess_clients):
    if request.param == "docker":
        return docker_clients
    else:
        return subprocess_clients



import pytest

from substra.sdk.backends.remote import request_formatter


@pytest.mark.parametrize(
    "raw,formatted",
    [
        ({"foo": ["bar", "baz"]}, {"foo": "bar,baz"}),
        ({"foo": ["bar", "baz"], "bar": ["qux"]}, {"foo": "bar,baz", "bar": "qux"}),
        ({"foo": ["b ar ", " baz"]}, {"foo": "bar,baz"}),
        (
            {},
            {},
        ),
        ({"name": "bar,baz"}, {"match": "bar,baz"}),
        (
            {"metadata": [{"key": "epochs", "type": "is", "value": "10"}]},
            {"metadata": '[{"key":"epochs","type":"is","value":"10"}]'},
        ),
        (None, {}),
    ],
)
def test_format_search_filters_for_remote(raw, formatted):
    assert request_formatter.format_search_filters_for_remote(raw) == formatted


@pytest.mark.parametrize(
    "ordering, ascending, formatted",
    [
        ("creation_date", False, "-creation_date"),
        ("start_date", True, "start_date"),
    ],
)
def test_format_search_ordering_for_remote(ordering, ascending, formatted):
    assert request_formatter.format_search_ordering_for_remote(ordering, ascending) == formatted



from unittest import mock

import requests


def mock_response(response=None, status=200, headers=None, json_error=None):
    headers = headers or {}
    m = mock.MagicMock(spec=requests.Response)
    m.status_code = status
    m.headers = headers
    m.text = str(response)
    m.json = mock.MagicMock(return_value=response, headers=headers, side_effect=json_error)

    if status not in (200, 201):
        exception = requests.exceptions.HTTPError(str(status), response=m)
        m.raise_for_status = mock.MagicMock(side_effect=exception)

    return m


def mock_requests_responses(mocker, method, responses):
    return mocker.patch(
        f"substra.sdk.backends.remote.rest_client.requests.{method}",
        side_effect=responses,
    )


def mock_requests(mocker, method, response=None, status=200, headers=None, json_error=None):
    r = mock_response(response, status, headers, json_error)
    return mock_requests_responses(mocker, method, (r,))


def make_paginated_response(items):
    return {"count": len(items), "next": None, "previous": None, "results": items}



import pydantic
import pytest

import substra
from substra.sdk import models
from substra.sdk.exceptions import ComputePlanKeyFormatError

from .. import datastore
from ..utils import mock_requests


def test_add_dataset(client, dataset_query, mocker):
    m_post = mock_requests(mocker, "post", response=datastore.DATASET)
    m_get = mock_requests(mocker, "get", response=datastore.DATASET)
    key = client.add_dataset(dataset_query)
    response = client.get_dataset(key)

    assert response == models.Dataset(**datastore.DATASET)
    m_post.assert_called()
    m_get.assert_called()


def test_add_dataset_invalid_args(client, dataset_query, mocker):
    mock_requests(mocker, "post", response=datastore.DATASET)
    del dataset_query["data_opener"]

    with pytest.raises(pydantic.ValidationError):
        client.add_dataset(dataset_query)


def test_add_dataset_response_failure_500(client, dataset_query, mocker):
    mock_requests(mocker, "post", status=500)

    with pytest.raises(substra.sdk.exceptions.InternalServerError):
        client.add_dataset(dataset_query)


def test_add_dataset_409_success(client, dataset_query, mocker):
    mock_requests(mocker, "post", response={"key": datastore.DATASET["key"]}, status=409)
    mock_requests(mocker, "get", response=datastore.DATASET)

    key = client.add_dataset(dataset_query)

    assert key == datastore.DATASET["key"]


def test_add_function(client, function_query, mocker):
    m_post = mock_requests(mocker, "post", response=datastore.FUNCTION)
    m_get = mock_requests(mocker, "get", response=datastore.FUNCTION)
    key = client.add_function(function_query)
    response = client.get_function(key)

    assert response == models.Function(**datastore.FUNCTION)
    m_post.assert_called()
    m_get.assert_called()


def test_add_data_sample(client, data_sample_query, mocker):
    server_response = [{"key": "42"}]
    m = mock_requests(mocker, "post", response=server_response)
    response = client.add_data_sample(data_sample_query)

    assert response == server_response[0]["key"]
    m.assert_called()


def test_add_data_sample_already_exists(client, data_sample_query, mocker):
    m = mock_requests(mocker, "post", response=[{"key": "42"}], status=409)
    response = client.add_data_sample(data_sample_query)

    assert response == "42"
    m.assert_called()


# We try to add multiple data samples instead of a single one
def test_add_data_sample_with_paths(client, data_samples_query):
    with pytest.raises(ValueError):
        client.add_data_sample(data_samples_query)


def test_add_data_samples(client, data_samples_query, mocker):
    server_response = [{"key": "42"}]
    m = mock_requests(mocker, "post", response=server_response)
    response = client.add_data_samples(data_samples_query)

    assert response == ["42"]
    m.assert_called()


# We try to add a single data sample instead of multiple ones
def test_add_data_samples_with_path(client, data_sample_query):
    with pytest.raises(ValueError):
        client.add_data_samples(data_sample_query)


def test_add_compute_plan_wrong_key_format(client):
    with pytest.raises(ComputePlanKeyFormatError):
        data = {"key": "wrong_format", "name": "A perfectly valid name"}
        client.add_compute_plan(data)



import contextlib
import datetime
import json
import os
import pathlib
import shutil
import string
import uuid
from enum import Enum
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

from substra.sdk import exceptions
from substra.sdk import fs
from substra.sdk import models
from substra.sdk import schemas
from substra.sdk.backends.local import dal
from substra.sdk.backends.local import models as models_local
from substra.sdk.backends.local.compute import spawner
from substra.sdk.backends.local.compute.spawner import BaseSpawner
from substra.sdk.backends.local.compute.spawner.base import VOLUME_CLI_ARGS
from substra.sdk.backends.local.compute.spawner.base import VOLUME_INPUTS
from substra.sdk.backends.local.compute.spawner.base import VOLUME_OUTPUTS

TPL_VOLUME_INPUTS = "${" + VOLUME_INPUTS + "}"
TPL_VOLUME_OUTPUTS = "${" + VOLUME_OUTPUTS + "}"


class TaskResource(dict):
    def __init__(self, id: str, value: str, multiple: bool):
        super().__init__(self, id=id, value=value, multiple=multiple)


class Filenames(str, Enum):
    OPENER = "opener.py"  # The name must end in .py for substra-tools module loading


def _mkdir(path, delete_if_exists=False):
    """Make directory (recursive)."""
    path = Path(path)
    if path.exists():
        if not delete_if_exists:
            return path
        shutil.rmtree(path)
    path.mkdir(parents=True)
    return path


def _generate_filename():
    return str(uuid.uuid4())


class Worker:
    """ML Worker."""

    def __init__(
        self,
        db: dal.DataAccess,
        local_worker_dir: pathlib.Path,
        debug_spawner: Type[BaseSpawner],
    ):
        self._local_worker_dir = local_worker_dir
        self._db = db
        self._spawner = spawner.get(name=debug_spawner, local_worker_dir=self._local_worker_dir)

    @contextlib.contextmanager
    def _context(self, key):
        try:
            tmp_dir = _mkdir(os.path.join(self._local_worker_dir, key))
            yield Path(tmp_dir)
        finally:
            # delete task working directory
            shutil.rmtree(tmp_dir, ignore_errors=True)

    def _save_cp_performances_as_json(self, compute_plan_key: str, path: Path) -> None:
        """Dump a json file containing the performances of the given compute plan in the given path."""

        performances = self._db.get_performances(compute_plan_key)

        with (path).open("w", encoding="UTF-8") as json_file:
            json.dump(performances.model_dump(), json_file, default=str)

    def _get_asset_unknown_type(self, asset_key, possible_types: List[schemas.Type]) -> Tuple[Any, schemas.Type]:
        for asset_type in possible_types:
            try:
                asset = self._db.get(asset_type, asset_key)
                break
            except exceptions.NotFound:
                pass
        if asset is None:
            raise exceptions.NotFound(f"Wrong pk {asset_key}", 404)
        return asset, asset_type

    def _update_cp(self, compute_plan: models.ComputePlan, update_live_performances: bool):
        compute_plan.done_count += 1
        compute_plan.todo_count -= 1
        if compute_plan.done_count == compute_plan.task_count:
            compute_plan.status = models.ComputePlanStatus.done
            compute_plan.end_date = datetime.datetime.now()
            compute_plan.estimated_end_date = compute_plan.end_date

            compute_plan.duration = int((compute_plan.end_date - compute_plan.start_date).total_seconds())

        # save live performances
        if update_live_performances:
            live_perf_path = Path(self._local_worker_dir / "live_performances" / compute_plan.key / "performances.json")
            _mkdir(live_perf_path.parent)
            self._save_cp_performances_as_json(compute_plan.key, live_perf_path)

    def _get_cmd_template_inputs_outputs(
        self,
        task,
        cmd_line_inputs,
        output_id_filename,
    ):
        task_properties = {"rank": task.rank}
        command_template = ["--task-properties", json.dumps(task_properties, default=str)]

        cmd_line_outputs: List[TaskResource] = [
            TaskResource(id=output_id, value=f"{TPL_VOLUME_OUTPUTS}/{filename}", multiple=False)
            for output_id, filename in output_id_filename.items()
        ]

        command_template += [
            "--inputs",
            json.dumps(cmd_line_inputs, default=str),
            "--outputs",
            json.dumps(cmd_line_outputs, default=str),
        ]
        return command_template

    def _prepare_artifact_input(self, task_input, task_key, input_volume, multiple):
        outputs = self._db.list(
            schemas.Type.OutputAsset,
            {"compute_task_key": task_input.parent_task_key, "identifier": task_input.parent_task_output_identifier},
        )
        assert len(outputs) == 1, "The input should be unique"
        output = outputs[0]
        self._db.add(
            models_local.InputAssetLocal(
                compute_task_key=task_key, kind=output.kind, identifier=output.identifier, asset=output.asset
            )
        )
        assert output.kind == schemas.AssetKind.model, "The task_input value must be an artifact, not a performance"
        filename = _generate_filename()
        path_to_input = input_volume / filename
        Path(output.asset.address.storage_address).link_to(path_to_input)

        return TaskResource(id=task_input.identifier, value=f"{TPL_VOLUME_INPUTS}/{filename}", multiple=multiple)

    def _prepare_dataset_input(
        self, dataset: models.Dataset, task_input: models.InputRef, input_volume: str, multiple: bool
    ):
        path_to_opener = input_volume / Filenames.OPENER.value
        Path(dataset.opener.storage_address).link_to(path_to_opener)
        return TaskResource(
            id=task_input.identifier,
            value=f"{TPL_VOLUME_INPUTS}/{Filenames.OPENER.value}",
            multiple=multiple,
        )

    def _prepare_datasample_input(
        self, datasample_input_refs: List[models.InputRef], datasamples: List[models.DataSample], multiple: bool
    ) -> Tuple[List[TaskResource], Dict[str, str]]:
        task_resources = []
        data_sample_paths = dict()
        for datasample_input, datasample in zip(datasample_input_refs, datasamples):
            datasample_path_arg = f"{TPL_VOLUME_INPUTS}/{datasample_input.asset_key}"
            task_resources.append(
                TaskResource(id=datasample_input.identifier, value=str(datasample_path_arg), multiple=multiple)
            )
            data_sample_paths[datasample.key] = datasample.path
        return task_resources, data_sample_paths

    def _prepare_datasamples_inputs_and_paths(
        self,
        task,
        dataset: models.Dataset,
        datasamples: List[models.DataSample],
        datasample_input_refs: List[models.InputRef],
        multiple: bool,
    ) -> Tuple[str, List[TaskResource], Optional[Dict[str, str]]]:
        command_template = []
        datasample_task_resources = []
        data_sample_paths = None

        if len(datasamples) == 0:
            pass
        elif dataset is not None and not self._db.is_local(dataset.key, schemas.Type.Dataset):
            # Hybrid mode
            command_template += ["--fake-data"]
            command_template += ["--n-fake-samples", len(datasample_input_refs)]
        else:
            datasample_task_resources, data_sample_paths = self._prepare_datasample_input(
                datasample_input_refs=datasample_input_refs, datasamples=datasamples, multiple=multiple
            )

        return command_template, datasample_task_resources, data_sample_paths

    def _save_output(
        self,
        *,
        task,
        function_output: models.FunctionOutput,
        output_id_filename: Dict[str, str],
        output_volume: str,
        function_key: str,
    ) -> bool:
        update_live_performances = False
        if function_output.multiple:
            raise NotImplementedError("Multiple output value is not supported yet.")
        filename = output_id_filename[function_output.identifier]
        tmp_path = output_volume / filename
        output_dir = _mkdir(self._local_worker_dir / "outputs" / task.key)
        output_path = output_dir / filename
        shutil.copy(tmp_path, output_path)

        if function_output.kind == schemas.AssetKind.performance:
            update_live_performances = True
            value = json.loads(output_path.read_text())["all"]
        elif function_output.kind == schemas.AssetKind.model:
            value = models.OutModel(
                key=str(uuid.uuid4()),
                compute_task_key=task.key,
                address=models.InModel(
                    checksum=fs.hash_file(output_path),
                    storage_address=output_path,
                ),
                creation_date=datetime.datetime.now(),
                owner=task.owner,
                permissions=task.outputs[function_output.identifier].permissions,
            )
            self._db.add(value)
        else:
            raise ValueError(f"This asset kind is not supported for function output: {function_output.kind}")
        output_asset = models_local.OutputAssetLocal(
            kind=function_output.kind,
            identifier=function_output.identifier,
            asset=value,
            compute_task_key=task.key,
        )
        self._db.add(output_asset)
        return update_live_performances

    def schedule_task(self, task: models.Task):
        """Execute the task

        Args:
            task: Task to execute
        """
        with self._context(task.key) as task_dir:
            task.status = models.Status.doing
            task.start_date = datetime.datetime.now()

            function = self._db.get_with_files(schemas.Type.Function, task.function.key)
            input_multiplicity = {i.identifier: i.multiple for i in function.inputs}
            compute_plan = self._db.get(schemas.Type.ComputePlan, task.compute_plan_key)

            command_template = []

            volumes = {
                VOLUME_INPUTS: _mkdir(task_dir / "inputs"),
                VOLUME_OUTPUTS: _mkdir(task_dir / "outputs"),
                VOLUME_CLI_ARGS: _mkdir(task_dir / "cli-args"),
            }

            cmd_line_inputs: List[TaskResource] = []

            dataset: Optional[models.Dataset] = None
            data_sample_paths: Optional[Dict[str, str]] = None

            datasample_input_refs: List[models.InputRef] = []
            datasamples: List[models.DataSample] = []

            addable_asset: Optional[Union[models.DataSample, models.Dataset]] = None
            # Prepare inputs
            for task_input in task.inputs:
                multiple = input_multiplicity[task_input.identifier]

                if task_input.parent_task_key is not None:
                    task_resource = self._prepare_artifact_input(
                        task_input=task_input,
                        task_key=task.key,
                        input_volume=volumes[VOLUME_INPUTS],
                        multiple=multiple,
                    )
                    cmd_line_inputs.append(task_resource)
                else:
                    asset, asset_type = self._get_asset_unknown_type(
                        asset_key=task_input.asset_key, possible_types=[schemas.Type.DataSample, schemas.Type.Dataset]
                    )

                    if asset_type == schemas.Type.DataSample:
                        # This is necessary because of the hybrid mode
                        # Otherwise could process the task_input here
                        datasample_input_refs.append(task_input)
                        datasamples.append(asset)

                        # In hybrid mode, we use DataSample assets with a path equals to `None` as it created as fake
                        # data in `_prepare_datasamples_inputs_and_paths`, but if we add these Data samples to the DB,
                        # during the execution substra will try to copy file in these folders and causes an error as
                        # these are not paths.
                        if asset.path:
                            addable_asset = asset
                    elif asset_type == schemas.Type.Dataset:
                        dataset = self._db.get_with_files(schemas.Type.Dataset, task_input.asset_key)
                        cmd_line_inputs.append(
                            self._prepare_dataset_input(
                                dataset=dataset,
                                task_input=task_input,
                                input_volume=volumes[VOLUME_INPUTS],
                                multiple=multiple,
                            )
                        )
                        addable_asset = dataset

                    if addable_asset:
                        input_asset = models_local.InputAssetLocal(
                            compute_task_key=task.key,
                            kind=asset_type.to_asset_kind(),
                            identifier=task_input.identifier,
                            asset=addable_asset,
                        )
                        self._db.add(input_asset)

                    addable_asset = None

            (
                datasample_cmd_template,
                datasample_task_resources,
                data_sample_paths,
            ) = self._prepare_datasamples_inputs_and_paths(
                task=task,
                dataset=dataset,
                datasamples=datasamples,
                datasample_input_refs=datasample_input_refs,
                multiple=input_multiplicity.get(schemas.StaticInputIdentifier.datasamples),
            )
            command_template += datasample_cmd_template
            cmd_line_inputs.extend(datasample_task_resources)

            # Prepare the outputs
            output_id_filename: Dict[str, str] = {output: _generate_filename() for output in task.outputs}

            command_template += self._get_cmd_template_inputs_outputs(
                task=task,
                cmd_line_inputs=cmd_line_inputs,
                output_id_filename=output_id_filename,
            )
            command_template += ["--log-level", "warning"]

            # Task execution
            container_name = f"function-{function.function.checksum}"
            self._spawner.spawn(
                container_name,
                str(function.function.storage_address),
                command_args_tpl=[string.Template(str(part)) for part in command_template],
                local_volumes=volumes,
                data_sample_paths=data_sample_paths,
                envs=None,
            )

            # Save the outputs
            update_live_performances = False
            for function_output in function.outputs:
                update_live_performances = self._save_output(
                    task=task,
                    function_output=function_output,
                    output_id_filename=output_id_filename,
                    output_volume=volumes[VOLUME_OUTPUTS],
                    function_key=function.key,
                )

            # Set status
            task.status = models.Status.done
            task.end_date = datetime.datetime.now()

            self._update_cp(compute_plan=compute_plan, update_live_performances=update_live_performances)



from tests import mocked_requests


def test_cancel_compute_plan(client, mocker):
    m = mocked_requests.cancel_compute_plan(mocker)

    response = client.cancel_compute_plan("magic-key")

    assert response is None
    m.assert_called()



import pytest

import substra
from substra.sdk.schemas import FunctionSpec

from . import data_factory
from .fl_interface import FLFunctionInputs
from .fl_interface import FLFunctionOutputs
from .fl_interface import FunctionCategory


def pytest_configure(config):
    config.addinivalue_line(
        "markers",
        "slow: marks tests as slow (deselect with '-m \"not slow\"')",
    )


@pytest.fixture
def client(tmpdir):
    c = substra.Client(url="http://foo.io", backend_type="remote", token="foo")
    return c


@pytest.fixture
def workdir(tmp_path):
    d = tmp_path / "substra-cli"
    d.mkdir()
    return d


@pytest.fixture
def dataset_query(tmpdir):
    opener_path = tmpdir / "opener.py"
    opener_path.write_text("raise ValueError()", encoding="utf-8")

    desc_path = tmpdir / "description.md"
    desc_path.write_text("#Hello world", encoding="utf-8")

    return {
        "name": "dataset_name",
        "data_opener": str(opener_path),
        "type": "images",
        "description": str(desc_path),
        "permissions": {
            "public": True,
            "authorized_ids": [],
        },
        "logs_permission": {
            "public": True,
            "authorized_ids": [],
        },
    }


@pytest.fixture
def metric_query(tmpdir):
    metrics_path = tmpdir / "metrics.zip"
    metrics_path.write_text("foo archive", encoding="utf-8")

    desc_path = tmpdir / "description.md"
    desc_path.write_text("#Hello world", encoding="utf-8")

    return {
        "name": "metrics_name",
        "file": str(metrics_path),
        "description": str(desc_path),
        "permissions": {
            "public": True,
            "authorized_ids": [],
        },
    }


@pytest.fixture
def function_query(tmpdir):
    function_file_path = tmpdir / "function.tar.gz"
    function_file_path.write(b"tar gz archive")

    desc_path = tmpdir / "description.md"
    desc_path.write_text("#Hello world", encoding="utf-8")

    function_category = FunctionCategory.simple

    return FunctionSpec(
        name="function_name",
        inputs=FLFunctionInputs[function_category],
        outputs=FLFunctionOutputs[function_category],
        description=str(desc_path),
        file=str(function_file_path),
        permissions={
            "public": True,
            "authorized_ids": [],
        },
    )


@pytest.fixture
def data_sample_query(tmpdir):
    data_sample_dir_path = tmpdir / "data_sample_0"
    data_sample_file_path = data_sample_dir_path / "data.txt"
    data_sample_file_path.write_text("Hello world 0", encoding="utf-8", ensure=True)

    return {
        "path": str(data_sample_dir_path),
        "data_manager_keys": ["42"],
    }


@pytest.fixture
def data_samples_query(tmpdir):
    nb = 3
    paths = []
    for i in range(nb):
        data_sample_dir_path = tmpdir / f"data_sample_{i}"
        data_sample_file_path = data_sample_dir_path / "data.txt"
        data_sample_file_path.write_text(f"Hello world {i}", encoding="utf-8", ensure=True)

        paths.append(str(data_sample_dir_path))

    return {
        "paths": paths,
        "data_manager_keys": ["42"],
    }


@pytest.fixture(scope="session")
def asset_factory():
    return data_factory.AssetsFactory("test_debug")


@pytest.fixture()
def data_sample(asset_factory):
    return asset_factory.create_data_sample()



import logging
import pathlib
import shutil
import string
import tempfile
import typing

import docker

from substra.sdk.archive import uncompress
from substra.sdk.backends.local.compute.spawner.base import VOLUME_CLI_ARGS
from substra.sdk.backends.local.compute.spawner.base import VOLUME_INPUTS
from substra.sdk.backends.local.compute.spawner.base import VOLUME_OUTPUTS
from substra.sdk.backends.local.compute.spawner.base import BaseSpawner
from substra.sdk.backends.local.compute.spawner.base import BuildError
from substra.sdk.backends.local.compute.spawner.base import ExecutionError
from substra.sdk.backends.local.compute.spawner.base import write_command_args_file

logger = logging.getLogger(__name__)

ROOT_DIR = "/substra_internal"
DOCKER_VOLUMES = {
    VOLUME_INPUTS: {"bind": f"{ROOT_DIR}/inputs", "mode": "ro"},
    VOLUME_OUTPUTS: {"bind": f"{ROOT_DIR}/outputs", "mode": "rw"},
    VOLUME_CLI_ARGS: {"bind": f"{ROOT_DIR}/cli-args", "mode": "rw"},
}


def _copy_data_samples(data_sample_paths: typing.Dict[str, pathlib.Path], dest_dir: str):
    """Move the data samples to the data directory.

    We copy the data samples even though it is slow because:

    - symbolic links do not work with Docker container volumes
    - hard links cannot be created across partitions
    - mounting each data sample as its own volume causes permission errors

    Args:
        data_sample_paths (typing.Dict[str, pathlib.Path]): Paths to the samples
        dest_dir (str): Temp data directory
    """
    # Check if there are already data samples in the dest dir (testtasks are executed in 2 parts)
    sample_key = next(iter(data_sample_paths))
    if (pathlib.Path(dest_dir) / sample_key).exists():
        return
    # copy the whole tree
    for sample_key, sample_path in data_sample_paths.items():
        dest_path = pathlib.Path(dest_dir) / sample_key
        shutil.copytree(sample_path, dest_path)


class Docker(BaseSpawner):
    """Wrapper around docker daemon to execute a command in a container."""

    def __init__(self, local_worker_dir: pathlib.Path):
        try:
            self._docker = docker.from_env()
        except docker.errors.DockerException as e:
            raise ConnectionError(
                "Couldn't get the Docker client from environment variables. "
                "Is your Docker server running ?\n"
                "Docker error : {0}".format(e)
            )
        super().__init__(local_worker_dir=local_worker_dir)

    def _build_docker_image(self, name: str, archive_path: pathlib.Path):
        """Spawn a docker container (blocking)."""
        with tempfile.TemporaryDirectory(dir=self._local_worker_dir) as tmpdir:
            image_exists = False
            try:
                self._docker.images.get(name=name)
                image_exists = True
            except docker.errors.ImageNotFound:
                pass

            if not image_exists:
                try:
                    logger.debug("Did not find the Docker image %s - building it", name)
                    uncompress(archive_path, tmpdir)
                    self._docker.images.build(path=tmpdir, tag=name, rm=True)
                except docker.errors.BuildError as exc:
                    log = ""
                    for line in exc.build_log:
                        if "stream" in line:
                            log += line["stream"].strip()
                    logger.error(log)
                    raise BuildError(log)

    def spawn(
        self,
        name: str,
        archive_path: pathlib.Path,
        command_args_tpl: typing.List[string.Template],
        data_sample_paths: typing.Optional[typing.Dict[str, pathlib.Path]],
        local_volumes: typing.Optional[dict],
        envs: typing.Optional[typing.List[str]],
    ):
        """Build the docker image, copy the data samples then spawn a Docker container
        and execute the task.
        """
        self._build_docker_image(name=name, archive_path=archive_path)

        # format the command to replace each occurrence of a DOCKER_VOLUMES's key
        # by its "bind" value
        volumes_format = {volume_name: volume_path["bind"] for volume_name, volume_path in DOCKER_VOLUMES.items()}
        command_args = [tpl.substitute(**volumes_format) for tpl in command_args_tpl]

        if data_sample_paths is not None and len(data_sample_paths) > 0:
            _copy_data_samples(data_sample_paths, local_volumes[VOLUME_INPUTS])

        args_filename = "arguments.txt"
        args_path_local = pathlib.Path(local_volumes[VOLUME_CLI_ARGS]) / args_filename
        # Pathlib is incompatible here for Windows, as it would create a "WindowsPath"
        args_path_docker = DOCKER_VOLUMES[VOLUME_CLI_ARGS]["bind"] + "/" + args_filename
        write_command_args_file(args_path_local, command_args)

        # create the volumes dict for docker by binding the local_volumes and the DOCKER_VOLUME
        volumes_docker = {
            volume_path: DOCKER_VOLUMES[volume_name] for volume_name, volume_path in local_volumes.items()
        }

        container = self._docker.containers.run(
            name,
            command=f"@{args_path_docker}",
            volumes=volumes_docker or {},
            environment=envs,
            remove=False,
            detach=True,
            tty=True,
            stdin_open=True,
            shm_size="8G",
        )

        execution_logs = []
        for line in container.logs(stream=True, stdout=True, stderr=True):
            execution_logs.append(line.decode("utf-8"))

        r = container.wait()
        execution_logs_str = "".join(execution_logs)
        exit_code = r["StatusCode"]
        if exit_code != 0:
            logger.error("\n\nExecution logs: %s", execution_logs_str)
            raise ExecutionError(f"Container '{name}' exited with status code '{exit_code}'")

        container.remove()



from contextlib import nullcontext as does_not_raise

import pytest

from substra.sdk import exceptions
from substra.sdk.models import ComputePlanStatus
from substra.sdk.models import Status
from substra.sdk.models import TaskErrorType

from .. import datastore
from ..utils import mock_requests


def _param_name_maker(arg):
    if isinstance(arg, str):
        return arg
    else:
        return ""


@pytest.mark.parametrize(
    ("asset_dict", "function_name", "status", "expectation"),
    [
        (datastore.TRAINTASK, "wait_task", Status.done, does_not_raise()),
        (datastore.TRAINTASK, "wait_task", Status.canceled, pytest.raises(exceptions.FutureFailureError)),
        (datastore.COMPUTE_PLAN, "wait_compute_plan", ComputePlanStatus.done, does_not_raise()),
        (
            datastore.COMPUTE_PLAN,
            "wait_compute_plan",
            ComputePlanStatus.failed,
            pytest.raises(exceptions.FutureFailureError),
        ),
        (
            datastore.COMPUTE_PLAN,
            "wait_compute_plan",
            ComputePlanStatus.canceled,
            pytest.raises(exceptions.FutureFailureError),
        ),
    ],
    ids=_param_name_maker,
)
def test_wait(client, mocker, asset_dict, function_name, status, expectation):
    item = {**asset_dict, "status": status}
    mock_requests(mocker, "get", item)
    function = getattr(client, function_name)
    with expectation:
        function(key=item["key"])


def test_wait_task_failed(client, mocker):
    # We need an error type to stop the iteration
    item = {**datastore.TRAINTASK, "status": Status.failed, "error_type": TaskErrorType.internal}
    mock_requests(mocker, "get", item)
    with pytest.raises(exceptions.FutureFailureError):
        client.wait_task(key=item["key"])


@pytest.mark.parametrize(
    ("asset_dict", "function_name", "status"),
    [
        (datastore.TRAINTASK, "wait_task", Status.todo),
        (datastore.COMPUTE_PLAN, "wait_compute_plan", ComputePlanStatus.todo),
    ],
    ids=_param_name_maker,
)
def test_wait_timeout(client, mocker, asset_dict, function_name, status):
    item = {**asset_dict, "status": status}
    mock_requests(mocker, "get", item)
    function = getattr(client, function_name)
    with pytest.raises(exceptions.FutureTimeoutError):
        # mock_requests returns only once and timeout=0 is falsy, so setting a microscopic duration
        function(key=item["key"], timeout=1e-10)



import pytest

from substra.sdk import models
from substra.sdk import schemas

from .. import datastore
from ..utils import mock_requests


@pytest.mark.parametrize(
    "asset_type",
    ["dataset", "function", "compute_plan"],
)
def test_update_asset(asset_type, client, mocker):
    update_method = getattr(client, f"update_{asset_type}")
    get_method = getattr(client, f"get_{asset_type}")

    item = getattr(datastore, asset_type.upper())
    name_update = {"name": "New name"}
    updated_item = {**item, **name_update}

    m_put = mock_requests(mocker, "put")
    m_get = mock_requests(mocker, "get", response=updated_item)

    update_method("magic-key", name_update["name"])
    response = get_method("magic-key")

    assert response == models.SCHEMA_TO_MODEL[schemas.Type(asset_type)](**updated_item)
    m_put.assert_called()
    m_get.assert_called()


def test_add_compute_plan_tasks(client, mocker):
    item = datastore.COMPUTE_PLAN
    m = mock_requests(mocker, "post", response=item)
    m_get = mock_requests(mocker, "get", response=datastore.COMPUTE_PLAN)

    response = client.add_compute_plan_tasks("foo", {})

    assert response == models.ComputePlan(**item)
    m.assert_called
    m_get.assert_called


def test_add_compute_plan_tasks_with_schema(client, mocker):
    item = datastore.COMPUTE_PLAN
    m = mock_requests(mocker, "post", response=item)
    m_get = mock_requests(mocker, "get", response=datastore.COMPUTE_PLAN)

    response = client.add_compute_plan_tasks("foo", schemas.UpdateComputePlanTasksSpec(key="foo"))

    assert response == models.ComputePlan(**item)
    m.assert_called
    m_get.assert_called



import os
import shutil
import tarfile
import tempfile

import pytest

from substra.sdk.archive import _untar
from substra.sdk.archive import _unzip
from substra.sdk.archive import tarsafe
from substra.sdk.archive.safezip import ZipFile


class TestsZipFile:
    # This zip file was specifically crafted and contains empty files named:
    # foo/bar
    # ../foo/bar
    # ../../foo/bar
    # ../../../foo/bar
    TRAVERSAL_ZIP = os.path.join(os.path.dirname(__file__), "data", "traversal.zip")

    # This zip file was specifically crafted and contains:
    # bar
    # foo -> bar (symlink)
    SYMLINK_ZIP = os.path.join(os.path.dirname(__file__), "data", "symlink.zip")

    def test_raise_on_path_traversal(self):
        zf = ZipFile(self.TRAVERSAL_ZIP, "r")
        with pytest.raises(Exception) as exc:
            zf.extractall(tempfile.gettempdir())

        assert "Attempted directory traversal" in str(exc.value)

    def test_raise_on_symlink(self):
        zf = ZipFile(self.SYMLINK_ZIP, "r")
        with pytest.raises(Exception) as exc:
            zf.extractall(tempfile.gettempdir())

        assert "Unsupported symlink" in str(exc.value)

    def test_compress_uncompress_zip(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            path_testdir = os.path.join(tmpdirname, "testdir")
            os.makedirs(path_testdir)
            with open(os.path.join(path_testdir, "testfile.txt"), "w") as f:
                f.write("testtext")
            shutil.make_archive(os.path.join(tmpdirname, "test_archive"), "zip", root_dir=os.path.dirname(path_testdir))
            _unzip(os.path.join(tmpdirname, "test_archive.zip"), os.path.join(tmpdirname, "test_archive"))

            assert os.listdir(tmpdirname + "/test_archive/testdir") == ["testfile.txt"]


class TestsTarSafe:
    def test_compress_uncompress_tar(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            path_testdir = os.path.join(tmpdirname, "testdir")
            os.makedirs(path_testdir)
            with open(os.path.join(path_testdir, "testfile.txt"), "w") as f:
                f.write("testtext")
            path_tarfile = os.path.join(tmpdirname, "test_archive.tar")
            with tarsafe.open(path_tarfile, "w:gz") as tar:
                tar.add(path_testdir, arcname=os.path.basename(path_testdir))

            _untar(path_tarfile, os.path.join(tmpdirname, "test_archive"))

            assert os.listdir(tmpdirname + "/test_archive/testdir") == ["testfile.txt"]

    def test_raise_on_symlink(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            # create the following tree structure:
            # ./Dockerfile
            # ./foo
            #    ./Dockerfile -> ../Dockerfile

            filename = "Dockerfile"
            symlink_source = os.path.join(tmpdir, filename)
            with open(symlink_source, "w") as fp:
                fp.write("FROM bar")

            archive_root = os.path.join(tmpdir, "foo")
            os.mkdir(archive_root)
            os.symlink(symlink_source, os.path.join(archive_root, filename))

            # create a tar archive of the foo folder
            tarpath = os.path.join(tmpdir, "foo.tgz")
            with tarfile.open(tarpath, "w:gz") as tar:
                for root, _, files in os.walk(archive_root):
                    for file in files:
                        tar.add(os.path.join(root, file))

            with pytest.raises(tarsafe.TarSafeError) as error:
                with tarsafe.open(tarpath, "r") as tar:
                    tar.extractall()

                assert "Unsupported symlink" in str(error.exception)



import os
from unittest.mock import patch

import pytest

import substra
from substra.sdk import Client

from .. import datastore
from ..utils import mock_requests
from ..utils import mock_requests_responses
from ..utils import mock_response


@pytest.mark.parametrize(
    "asset_type",
    [
        ("dataset"),
        ("function"),
        ("model"),
    ],
)
def test_download_asset(asset_type, tmp_path, client, mocker):
    item = getattr(datastore, asset_type.upper())
    responses = [
        mock_response(item),  # metadata
        mock_response("foo"),  # data
    ]
    m = mock_requests_responses(mocker, "get", responses)

    method = getattr(client, f"download_{asset_type}")
    temp_file = method("foo", tmp_path)

    assert os.path.exists(temp_file)
    m.assert_called()


@pytest.mark.parametrize("asset_type", ["dataset", "function", "model", "logs"])
def test_download_asset_not_found(asset_type, tmp_path, client, mocker):
    m = mock_requests(mocker, "get", status=404)

    with pytest.raises(substra.sdk.exceptions.NotFound):
        method = getattr(client, f"download_{asset_type}")
        method("foo", tmp_path)

    assert m.call_count == 1


@pytest.mark.parametrize("asset_type", ["dataset", "function", "model"])
def test_download_content_not_found(asset_type, tmp_path, client, mocker):
    item = getattr(datastore, asset_type.upper())

    expected_call_count = 2
    responses = [
        mock_response(item),  # metadata
        mock_response("foo", status=404),  # description
    ]

    if asset_type == "model":
        responses = [responses[1]]  # No metadata for model download
        expected_call_count = 1

    m = mock_requests_responses(mocker, "get", responses)

    method = getattr(client, f"download_{asset_type}")

    with pytest.raises(substra.sdk.exceptions.NotFound):
        method("key", tmp_path)

    assert m.call_count == expected_call_count


@pytest.mark.parametrize(
    "asset_type, identifier",
    [
        ("TRAINTASK", "model"),
        ("AGGREGATETASK", "model"),
        ("COMPOSITE_TRAINTASK", "local"),
        ("COMPOSITE_TRAINTASK", "shared"),
    ],
)
@patch.object(Client, "download_model")
def test_download_model_from_task(fake_download_model, tmp_path, client, asset_type, identifier, mocker):
    item = getattr(datastore, f"{asset_type}_{identifier.upper()}_RESPONSE")
    responses = [
        mock_response(item),  # metadata
        mock_response("foo"),  # data
    ]

    m = mock_requests_responses(mocker, "get", responses)

    client.download_model_from_task("key", identifier, tmp_path)

    m.assert_called
    assert fake_download_model.call_count == 1


def test_download_logs(tmp_path, client, mocker):
    logs = b"Lorem ipsum dolor sit amet"
    task_key = "key"

    response = mock_response(logs)
    response.iter_content.return_value = [logs]

    m = mock_requests_responses(mocker, "get", [response])
    client.download_logs(task_key, tmp_path)

    m.assert_called_once()
    response.iter_content.assert_called_once()

    assert (tmp_path / f"task_logs_{task_key}.txt").read_bytes() == logs



import os
import zipfile

import pytest

import substra
from substra.sdk import exceptions
from substra.sdk import schemas
from substra.sdk import utils


def _unzip(fp, destination):
    with zipfile.ZipFile(fp, "r") as zipf:
        zipf.extractall(destination)


def test_zip_folder(tmp_path):
    # initialise dir to zip
    dir_to_zip = tmp_path / "dir"
    dir_to_zip.mkdir()

    file_items = [
        ("name0.txt", "content0"),
        ("dir1/name1.txt", "content1"),
        ("dir2/name2.txt", "content2"),
    ]

    for name, content in file_items:
        path = dir_to_zip / name
        path.parents[0].mkdir(exist_ok=True)
        path.write_text(content)

    for name, _ in file_items:
        path = dir_to_zip / name
        assert os.path.exists(str(path))

    # zip dir
    fp = utils.zip_folder_in_memory(str(dir_to_zip))
    assert fp

    # unzip dir
    destination_dir = tmp_path / "destination"
    destination_dir.mkdir()
    _unzip(fp, str(destination_dir))
    for name, content in file_items:
        path = destination_dir / name
        assert os.path.exists(str(path))
        assert path.read_text() == content


@pytest.mark.parametrize(
    "filters,expected,exception",
    [
        ("str", None, exceptions.FilterFormatError),
        ({}, None, exceptions.FilterFormatError),
        (
            [{"key": "foo", "type": "bar", "value": "baz"}],
            None,
            exceptions.FilterFormatError,
        ),
        ([{"key": "foo", "type": "is", "value": "baz"}, {}], None, exceptions.FilterFormatError),
        ([{"key": "foo", "type": "is", "value": "baz"}], None, None),
    ],
)
def test_check_metadata_search_filter(filters, expected, exception):
    if exception:
        with pytest.raises(exception):
            utils._check_metadata_search_filters(filters)
    else:
        assert utils._check_metadata_search_filters(filters) == expected


@pytest.mark.parametrize(
    "asset_type,filters,expected,exception",
    [
        (
            schemas.Type.ComputePlan,
            {"status": [substra.models.ComputePlanStatus.doing.value]},
            {"status": [substra.models.ComputePlanStatus.doing.value]},
            None,
        ),
        (
            schemas.Type.Task,
            {"status": [substra.models.Status.done.value]},
            {"status": [substra.models.Status.done.value]},
            None,
        ),
        (schemas.Type.Task, {"rank": [1]}, {"rank": ["1"]}, None),
        (schemas.Type.DataSample, ["wrong filter type"], None, exceptions.FilterFormatError),
        (schemas.Type.ComputePlan, {"name": ["list"]}, None, exceptions.FilterFormatError),
        (schemas.Type.Task, {"foo": "not allowed key"}, None, exceptions.NotAllowedFilterError),
        (
            schemas.Type.ComputePlan,
            {"name": "cp1", "key": ["key1", "key2"]},
            {"name": "cp1", "key": ["key1", "key2"]},
            None,
        ),
    ],
)
def test_check_and_format_search_filters(asset_type, filters, expected, exception):
    if exception:
        with pytest.raises(exception):
            utils.check_and_format_search_filters(asset_type, filters)
    else:
        assert utils.check_and_format_search_filters(asset_type, filters) == expected


@pytest.mark.parametrize(
    "ordering, exception",
    [
        ("creation_date", None),
        ("start_date", None),
        ("foo", exceptions.OrderingFormatError),
        (None, None),
    ],
)
def test_check_search_ordering(ordering, exception):
    if exception:
        with pytest.raises(exception):
            utils.check_search_ordering(ordering)
    else:
        utils.check_search_ordering(ordering)



import pytest

from substra.sdk import exceptions
from substra.sdk import models
from substra.sdk import schemas

from .. import datastore
from ..utils import make_paginated_response
from ..utils import mock_requests


@pytest.mark.parametrize(
    "asset_type",
    [
        "dataset",
        "function",
        "compute_plan",
        "data_sample",
        "model",
    ],
)
def test_list_asset(asset_type, client, mocker):
    item = getattr(datastore, asset_type.upper())
    method = getattr(client, f"list_{asset_type}")

    mocked_response = make_paginated_response([item])
    m = mock_requests(mocker, "get", response=mocked_response)

    response = method()

    assert response == [models.SCHEMA_TO_MODEL[schemas.Type(asset_type)](**item)]
    m.assert_called()


@pytest.mark.parametrize(
    "asset_type",
    [
        "predicttask",
        "testtask",
        "traintask",
        "aggregatetask",
        "composite_traintask",
    ],
)
def test_list_task(asset_type, client, mocker):
    item = getattr(datastore, asset_type.upper())

    mocked_response = make_paginated_response([item])
    m = mock_requests(mocker, "get", response=mocked_response)

    response = client.list_task()

    assert response == [models.Task(**item)]
    m.assert_called()


@pytest.mark.parametrize(
    "asset_type,filters",
    [
        ("dataset", {"permissions": ["foo", "bar"]}),
        ("function", {"owner": ["foo", "bar"]}),
        ("compute_plan", {"name": "foo"}),
        ("compute_plan", {"status": [models.ComputePlanStatus.done.value]}),
        ("model", {"owner": ["MyOrg1MSP"]}),
    ],
)
def test_list_asset_with_filters(asset_type, filters, client, mocker):
    item = getattr(datastore, asset_type.upper())
    method = getattr(client, f"list_{asset_type}")

    mocked_response = make_paginated_response([item])
    m = mock_requests(mocker, "get", response=mocked_response)

    response = method(filters)

    assert response == [models.SCHEMA_TO_MODEL[schemas.Type(asset_type)](**item)]
    m.assert_called()


@pytest.mark.parametrize(
    "filters",
    [
        {"rank": [1, 3]},
        {"key": ["foo", "bar"]},
        {"worker": ["foo", "bar"]},
        {"owner": ["foo", "bar"]},
    ],
)
def test_list_task_with_filters(filters, client, mocker):
    items = datastore.TASK_LIST

    mocked_response = make_paginated_response(items)
    m = mock_requests(mocker, "get", response=mocked_response)

    response = client.list_task(filters)

    assert response == [models.Task(**item) for item in items]
    m.assert_called()


def test_list_asset_with_filters_failure(client, mocker):
    items = [datastore.FUNCTION]
    m = mock_requests(mocker, "get", response=items)

    filters = {"foo"}
    with pytest.raises(exceptions.FilterFormatError) as exc_info:
        client.list_function(filters)

    m.assert_not_called()
    assert str(exc_info.value).startswith("Cannot load filters")


def test_list_compute_plan_with_ordering(client, mocker):
    item = datastore.COMPUTE_PLAN

    mocked_response = make_paginated_response([item])
    m = mock_requests(mocker, "get", response=mocked_response)

    order_by = "start_date"
    response = client.list_compute_plan(order_by=order_by)

    assert response == [models.SCHEMA_TO_MODEL[schemas.Type.ComputePlan](**item)]
    m.assert_called()


def test_list_task_with_ordering(client, mocker):
    items = datastore.TASK_LIST

    mocked_response = make_paginated_response(items)
    m = mock_requests(mocker, "get", response=mocked_response)

    order_by = "start_date"
    response = client.list_task(order_by=order_by)

    assert response == [models.Task(**item) for item in items]
    m.assert_called()


def test_list_asset_with_ordering_failure(client, mocker):
    items = [datastore.COMPUTE_PLAN]
    m = mock_requests(mocker, "get", response=items)

    order_by = "foo"
    with pytest.raises(exceptions.OrderingFormatError) as exc_info:
        client.list_compute_plan(order_by=order_by)

    m.assert_not_called()
    assert str(exc_info.value).startswith("Please review the documentation")



import string

from substra.sdk.backends.local.compute.spawner.subprocess import _get_command_args


def test_get_command_args_without_space():
    # check that it's not changing the path if there is no spaces
    function_name = "train"
    command = ["--opener-path", "${_VOLUME_OPENER}", "--compute-plan-path", "${_VOLUME_LOCAL}"]
    command_template = [string.Template(part) for part in command]
    local_volumes = {
        "_VOLUME_OPENER": "/a/path/without/any/space/opener.py",
        "_VOLUME_LOCAL": "/another/path/without/any/space",
    }

    py_commands = _get_command_args(function_name, command_template, local_volumes)

    valid_py_commands = [
        "--function-name",
        "train",
        "--opener-path",
        "/a/path/without/any/space/opener.py",
        "--compute-plan-path",
        "/another/path/without/any/space",
    ]

    assert py_commands == valid_py_commands


def test_get_command_args_with_spaces():
    # check that it's not splitting path with spaces in different arguments
    function_name = "train"
    command = ["--opener-path", "${_VOLUME_OPENER}", "--compute-plan-path", "${_VOLUME_LOCAL}"]
    command_template = [string.Template(part) for part in command]
    local_volumes = {
        "_VOLUME_OPENER": "/a/path with spaces/opener.py",
        "_VOLUME_LOCAL": "/another/path with spaces",
    }

    py_commands = _get_command_args(function_name, command_template, local_volumes)

    valid_py_commands = [
        "--function-name",
        "train",
        "--opener-path",
        "/a/path with spaces/opener.py",
        "--compute-plan-path",
        "/another/path with spaces",
    ]

    assert py_commands == valid_py_commands



from substra.sdk import exceptions
from substra.sdk import graph
from substra.sdk import schemas


def _insert_into_graph(task_graph, task_id, in_model_ids):
    if task_id in task_graph:
        raise exceptions.InvalidRequest("Two tasks cannot have the same id.", 400)
    task_graph[task_id] = in_model_ids


def get_dependency_graph(spec: schemas._BaseComputePlanSpec):
    """Get the task dependency graph and, for each type of task, a mapping table id/task."""
    task_graph = {}
    tasks = {}

    if spec.tasks:
        for task in spec.tasks:
            _insert_into_graph(
                task_graph=task_graph,
                task_id=task.task_id,
                in_model_ids=[
                    input_ref.parent_task_key for input_ref in (task.inputs or []) if input_ref.parent_task_key
                ],
            )
            tasks[task.task_id] = schemas.TaskSpec.from_compute_plan(
                compute_plan_key=spec.key,
                rank=None,
                spec=task,
            )
    return task_graph, tasks


def get_tasks(spec):
    """Returns compute plan tasks sorted by dependencies."""

    # Create the dependency graph and get the dict of tasks by id
    task_graph, tasks = get_dependency_graph(spec)

    already_created_ids = set()
    # Here we get the pre-existing tasks and assign them the minimal rank
    for dependencies in task_graph.values():
        for dependency_id in dependencies:
            if dependency_id not in task_graph:
                already_created_ids.add(dependency_id)

    # Compute the relative ranks of the new tasks (relatively to each other, these
    # are not their actual ranks in the compute plan)
    id_ranks = graph.compute_ranks(node_graph=task_graph, node_to_ignore=already_created_ids)

    # Sort the tasks by rank
    sorted_by_rank = sorted(id_ranks.items(), key=lambda item: item[1])

    return [tasks[task_id] for task_id, _ in sorted_by_rank]



import json

import pytest
import requests

from substra.sdk import exceptions
from substra.sdk.backends.remote import rest_client
from substra.sdk.client import Client

from .. import datastore
from ..utils import mock_requests
from ..utils import mock_requests_responses
from ..utils import mock_response

CONFIG = {
    "url": "http://foo.com",
    "insecure": False,
}

CONFIG_SECURE = {
    "url": "http://foo.com",
    "insecure": False,
}

CONFIG_INSECURE = {
    "url": "http://foo.com",
    "insecure": True,
}

CONFIGS = [CONFIG, CONFIG_SECURE, CONFIG_INSECURE]


def _client_from_config(config):
    return rest_client.Client(
        config["url"],
        config["insecure"],
        None,
    )


@pytest.mark.parametrize("config", CONFIGS)
def test_post_success(mocker, config):
    m = mock_requests(mocker, "post", response={})
    _client_from_config(config).add("http://foo", {})
    assert len(m.call_args_list) == 1


@pytest.mark.parametrize("config", CONFIGS)
def test_verify_login(mocker, config):
    """
    check "insecure" configuration results in endpoints being called with verify=False
    """
    m_post = mock_requests(mocker, "post", response={"id": "a", "token": "a", "expires_at": "3000-01-01T00:00:00Z"})
    m_delete = mock_requests(mocker, "delete", response={})

    c = _client_from_config(config)
    c.login("foo", "bar")
    c.logout()
    if config.get("insecure", None):
        assert m_post.call_args.kwargs["verify"] == False
        assert m_delete.call_args.kwargs["verify"] == False
    else:
        assert "verify" not in m_post.call_args.kwargs or m_post.call_args.kwargs["verify"]
        assert "verify" not in m_post.call_args.kwargs or m_delete.call_args.kwargs["verify"]


@pytest.mark.parametrize(
    "status_code, http_response, sdk_exception",
    [
        (400, {"detail": "Invalid Request"}, exceptions.InvalidRequest),
        (401, {"detail": "Invalid username/password"}, exceptions.AuthenticationError),
        (403, {"detail": "Unauthorized"}, exceptions.AuthorizationError),
        (404, {"detail": "Not Found"}, exceptions.NotFound),
        (408, {"key": "a-key"}, exceptions.RequestTimeout),
        (408, {}, exceptions.RequestTimeout),
        (500, "CRASH", exceptions.InternalServerError),
    ],
)
def test_request_http_errors(mocker, status_code, http_response, sdk_exception):
    m = mock_requests(mocker, "post", response=http_response, status=status_code)
    with pytest.raises(sdk_exception):
        _client_from_config(CONFIG).add("http://foo", {})
    assert len(m.call_args_list) == 1


def test_request_connection_error(mocker):
    mocker.patch(
        "substra.sdk.backends.remote.rest_client.requests.post", side_effect=requests.exceptions.ConnectionError
    )
    with pytest.raises(exceptions.ConnectionError):
        _client_from_config(CONFIG).add("foo", {})


def test_add_timeout_with_retry(mocker):
    asset_type = "traintask"
    responses = [
        mock_response(response={"key": "a-key"}, status=408),
        mock_response(response={"key": "a-key"}),
    ]
    m_post = mock_requests_responses(mocker, "post", responses)
    asset = _client_from_config(CONFIG).add(asset_type, retry_timeout=60)
    assert len(m_post.call_args_list) == 2
    assert asset == {"key": "a-key"}


def test_add_already_exist(mocker):
    asset_type = "traintask"
    m_post = mock_requests(mocker, "post", response={"key": "a-key"}, status=409)
    asset = _client_from_config(CONFIG).add(asset_type)
    assert len(m_post.call_args_list) == 1
    assert asset == {"key": "a-key"}


def test_add_wrong_url(mocker):
    """Check correct error is raised when wrong url with correct syntax is set."""
    error = json.decoder.JSONDecodeError("", "", 0)

    mock_requests(mocker, "post", status=200, json_error=error)
    test_client = Client(url="http://www.dummy.com", token="foo")
    with pytest.raises(exceptions.BadConfiguration) as e:
        test_client.login("test_client", "hehe")
    assert "Make sure that given url" in e.value.args[0]


def test_list_paginated(mocker):
    asset_type = "traintask"
    items = [datastore.TRAINTASK, datastore.TRAINTASK]
    responses = [
        mock_response(
            response={
                "count": len(items),
                "next": "http://foo.com/?page=2",
                "previous": None,
                "results": items[:1],
            },
            status=200,
        ),
        mock_response(
            response={
                "count": len(items),
                "next": None,
                "previous": "http://foo.com/?page=1",
                "results": items[1:],
            },
            status=200,
        ),
    ]
    m_get = mock_requests_responses(mocker, "get", responses)
    asset = _client_from_config(CONFIG).list(asset_type)
    assert len(asset) == len(items)
    assert len(m_get.call_args_list) == 2


def test_list_not_paginated(mocker):
    asset_type = "traintask"
    items = [datastore.TRAINTASK, datastore.TRAINTASK]
    m_get = mock_requests(
        mocker,
        "get",
        response={
            "count": len(items),
            "next": "http://foo.com/?page=2",
            "previous": None,
            "results": items[:1],
        },
        status=200,
    )
    asset = _client_from_config(CONFIG).list(asset_type, paginated=False)
    assert len(asset) != len(items)
    assert len(m_get.call_args_list) == 1



import pandas as pd
import pytest

import substra
from substra.sdk import models
from substra.sdk import schemas

from .. import datastore
from ..utils import mock_requests
from ..utils import mock_requests_responses
from ..utils import mock_response


@pytest.mark.parametrize(
    "asset_type",
    [
        "model",
        "dataset",
        "data_sample",
        "function",
        "compute_plan",
    ],
)
def test_get_asset(asset_type, client, mocker):
    item = getattr(datastore, asset_type.upper())
    method = getattr(client, f"get_{asset_type}")

    m = mock_requests(mocker, "get", response=item)

    response = method("magic-key")

    assert response == models.SCHEMA_TO_MODEL[schemas.Type(asset_type)](**item)
    m.assert_called()


@pytest.mark.parametrize(
    "asset_type",
    [
        "predicttask",
        "testtask",
        "traintask",
        "aggregatetask",
        "composite_traintask",
    ],
)
def test_get_task(asset_type, client, mocker):
    item = getattr(datastore, asset_type.upper())

    m = mock_requests(mocker, "get", response=item)

    response = client.get_task("magic-key")

    assert response == models.SCHEMA_TO_MODEL[schemas.Type.Task](**item)
    m.assert_called()


def test_get_asset_not_found(client, mocker):
    mock_requests(mocker, "get", status=404)

    with pytest.raises(substra.sdk.exceptions.NotFound):
        client.get_dataset("magic-key")


@pytest.mark.parametrize(
    "asset_type",
    [
        "dataset",
        "function",
        "compute_plan",
        "model",
    ],
)
def test_get_extra_field(asset_type, client, mocker):
    item = getattr(datastore, asset_type.upper())
    raw = getattr(datastore, asset_type.upper()).copy()
    raw["unknown_extra_field"] = "some value"

    method = getattr(client, f"get_{asset_type}")

    m = mock_requests(mocker, "get", response=raw)

    response = method("magic-key")

    assert response == models.SCHEMA_TO_MODEL[schemas.Type(asset_type)](**item)
    m.assert_called()


@pytest.mark.parametrize(
    "asset_type",
    [
        "predicttask",
        "testtask",
        "traintask",
        "aggregatetask",
        "composite_traintask",
    ],
)
def test_get_task_extra_field(asset_type, client, mocker):
    item = getattr(datastore, asset_type.upper())
    raw = getattr(datastore, asset_type.upper()).copy()
    raw["unknown_extra_field"] = "some value"

    m = mock_requests(mocker, "get", response=raw)

    response = client.get_task("magic-key")

    assert response == models.SCHEMA_TO_MODEL[schemas.Type.Task](**item)
    m.assert_called()


def test_get_logs(client, mocker):
    logs = "Lorem ipsum dolor sit amet"
    task_key = "key"

    responses = [mock_response(logs)]
    m = mock_requests_responses(mocker, "get", responses)
    result = client.get_logs(task_key)

    m.assert_called_once()
    assert result == logs


def test_get_performances(client, mocker):
    """Test the get_performances features, and test the immediate conversion to pandas DataFrame."""
    cp_item = datastore.COMPUTE_PLAN
    perf_item = datastore.COMPUTE_PLAN_PERF

    m = mock_requests_responses(mocker, "get", [mock_response(cp_item), mock_response(perf_item)])

    response = client.get_performances("magic-key")
    results = response.model_dump()

    df = pd.DataFrame(results)
    assert list(df.columns) == list(results.keys())
    assert all(len(v) == df.shape[0] for v in results.values())
    assert m.call_count == 2



import pytest

import substra

from .. import datastore
from ..utils import mock_requests
from ..utils import mock_requests_responses
from ..utils import mock_response


@pytest.mark.parametrize("asset_type", ["dataset", "function"])
def test_describe_asset(asset_type, client, mocker):
    item = getattr(datastore, asset_type.upper())
    responses = [
        mock_response(item),  # metadata
        mock_response("foo"),  # data
    ]
    m = mock_requests_responses(mocker, "get", responses)

    method = getattr(client, f"describe_{asset_type}")
    response = method("magic-key")

    assert response == "foo"
    m.assert_called()


@pytest.mark.parametrize("asset_type", ["dataset", "function"])
def test_describe_asset_not_found(asset_type, client, mocker):
    m = mock_requests(mocker, "get", status=404)

    with pytest.raises(substra.sdk.exceptions.NotFound):
        method = getattr(client, f"describe_{asset_type}")
        method("foo")

    assert m.call_count == 1


@pytest.mark.parametrize("asset_type", ["dataset", "function"])
def test_describe_description_not_found(asset_type, client, mocker):
    item = getattr(datastore, asset_type.upper())
    responses = [
        mock_response(item),  # metadata
        mock_response("foo", 404),  # data
    ]
    m = mock_requests_responses(mocker, "get", responses)

    method = getattr(client, f"describe_{asset_type}")

    with pytest.raises(substra.sdk.exceptions.NotFound):
        method("key")

    assert m.call_count == 2



from substra.sdk import schemas
from substra.sdk.backends.local.backend import Local
from substra.sdk.backends.remote.backend import Remote

_BACKEND_CHOICES = {
    schemas.BackendType.REMOTE: Remote,
    schemas.BackendType.LOCAL_DOCKER: Local,
    schemas.BackendType.LOCAL_SUBPROCESS: Local,
}


def get(name, *args, **kwargs):
    return _BACKEND_CHOICES[name](*args, **kwargs, backend_type=name)



import os

from substra.sdk.hasher import Hasher

_BLOCK_SIZE = 64 * 1024


def hash_file(path):
    """Hash a file."""
    hasher = Hasher()

    with open(path, "rb") as fp:
        while True:
            data = fp.read(_BLOCK_SIZE)
            if not data:
                break
            hasher.update(data)
    return hasher.compute()


def hash_directory(path, followlinks=False):
    """Hash a directory."""

    if not os.path.isdir(path):
        raise TypeError(f"{path} is not a directory.")

    hash_values = []
    for root, dirs, files in os.walk(path, topdown=True, followlinks=followlinks):
        dirs.sort()
        files.sort()

        for fname in files:
            hash_values.append(hash_file(os.path.join(root, fname)))

    return Hasher(values=sorted(hash_values)).compute()



import json
import logging
import time
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from typing import Dict
from typing import List
from typing import Union

import requests

from substra.sdk import exceptions
from substra.sdk import schemas
from substra.sdk import utils
from substra.sdk.backends.remote import request_formatter

logger = logging.getLogger(__name__)


def _warn_of_session_expiration(expiration: str, minimum_log_level=logging.INFO) -> None:
    log_level = minimum_log_level
    try:
        duration = datetime.fromisoformat(expiration) - datetime.now(timezone.utc)
        if duration.days:
            duration_msg = f"{duration.days} days"
        else:
            duration_msg = f"{duration.seconds//3600} hours"
        expires_at = f"in {duration_msg} ({expiration})"
        if duration < timedelta(hours=4):
            log_level = max(log_level, logging.WARNING)
    except ValueError:  # Python 3.11 parses more formats than previous versions
        expires_at = expiration
    logger.log(log_level, f"Your session will expire {expires_at}")


class Client:
    """REST Client to communicate with Substra server."""

    @property
    def base_url(self) -> str:
        return self._base_url

    def __init__(self, url, insecure, token):
        self._default_kwargs = {
            "verify": not insecure,
        }
        self._headers = {
            "Authorization": f"Token {token}",
            "Accept": "application/json;version=0.0",
        }
        if not url:
            raise exceptions.SDKException("url required to connect to the Substra server")
        self._base_url = url[:-1] if url.endswith("/") else url
        self._token = None  # filled in by self.login

    def login(self, username, password):
        # we do not use self._headers in order to avoid existing tokens to be sent alongside the
        # required Accept header
        if "Accept" not in self._headers:
            raise exceptions.SDKException("Cannot login: missing headers")

        headers = {
            "Accept": self._headers["Accept"],
        }
        data = {
            "username": username,
            "password": password,
        }
        try:
            r = requests.post(f"{self._base_url}/api-token-auth/", data=data, headers=headers, **self._default_kwargs)
            r.raise_for_status()
        except requests.exceptions.RequestException as e:
            if isinstance(e, requests.exceptions.HTTPError):
                if e.response.status_code in (400, 401):
                    raise exceptions.BadLoginException.from_request_exception(e)
                if (
                    e.response.status_code == 403
                    and getattr(e.response, "substra_identifier", None) == "implicit_login_disabled"
                ):
                    raise exceptions.UsernamePasswordLoginDisabledException.from_request_exception(e)
            raise exceptions.from_request_exception(e)

        try:
            rj = r.json()
            if not all([it in rj for it in ["token", "expires_at", "id"]]):
                raise exceptions.InvalidResponse(r, "Token object does not contain required fields")
            token = rj["token"]
            self._token = rj
        except json.decoder.JSONDecodeError:
            # sometimes requests seem to be fine, but the json is not being found
            # this might be if the url seems to be correct (in the syntax)
            # but it's not the right one
            raise exceptions.BadConfiguration(
                "Unable to get token from json response. " f"Make sure that given url: {self._base_url} is correct"
            )

        logger.info(
            "Logging in with username/password is discouraged; "
            "you should generate a token on the web UI instead: "
            "https://docs.substra.org/en/stable/documentation/api_tokens_generation.html"
        )
        _warn_of_session_expiration(self._token["expires_at"])
        self._headers["Authorization"] = f"Token {token}"

        return token

    def logout(self) -> None:
        if self._token is None:
            return
        try:
            r = requests.delete(
                f"{self._base_url}/active-api-tokens/",
                params={"id": self._token["id"]},
                headers=self._headers,
                **self._default_kwargs,
            )
            r.raise_for_status()
            self._token = None
            logger.info("Successfully logged out")
        except requests.exceptions.RequestException as e:
            logger.error(
                f"Could not end session {self._token['id']}"
                + (
                    f", got {e.response.status_code}: {e.response.text}"
                    if isinstance(e, requests.exceptions.HTTPError)
                    else f": {e}"
                )
            )
            _warn_of_session_expiration(self._token["expires_at"], minimum_log_level=logging.WARNING)
            # this isn't too much of an issue, the token will expire on its own

    def __request(self, request_name, url, **request_kwargs):
        """Base request helper."""

        if request_name == "get":
            fn = requests.get
        elif request_name == "post":
            fn = requests.post
        elif request_name == "put":
            fn = requests.put
        else:
            raise NotImplementedError

        # override default request arguments with input arguments
        kwargs = dict(self._default_kwargs)
        kwargs.update(request_kwargs)

        # rewind files so that they are properly sent in retries as well
        if "files" in kwargs:
            for file in kwargs["files"].values():
                file.seek(0)

        # do HTTP request and catch generic exceptions
        try:
            r = fn(url, headers=self._headers, **kwargs)
            r.raise_for_status()

        except requests.exceptions.RequestException as e:
            raise exceptions.from_request_exception(e)
        return r

    @utils.retry_on_exception(exceptions=(exceptions.GatewayUnavailable))
    def _request(self, request_name, url, **request_kwargs):
        """Wrapper to __request to emit a log for each HTTP request."""
        ts = time.time()
        error = None
        try:
            return self.__request(request_name, url, **request_kwargs)
        except Exception as e:
            error = e.__class__.__name__
            raise
        finally:
            te = time.time()
            elaps = (te - ts) * 1000
            logger.debug(f"{request_name} {url}: done in {elaps:.2f}ms error={error}")

    def request(
        self,
        request_name: str,
        asset_type: str,
        path: str = None,
        json_response: bool = True,
        paginated: bool = False,
        **request_kwargs,
    ):
        """
        Base request

        Args:
            request_name (str): http rest method invoked. e.g. "get", "post" or "put"
            asset_type (str): asset type. e.g. "function"
            path (str, optional): additional route e.g. "cp_key/perf" in /compute_plan/cp_key/perf. Defaults to None.
            json_response (bool, optional): whether the expected response is formatted in json. Defaults to True.
            paginated (bool, optional): whether the response would be paginated or not. Defaults to False.
            request_kwargs (dict): the args. The arg ["params"] is a dict with keys = attributes, values = list of
                values for this attribute. e.g. {name:[name1, name2]}

        Raises:
            exceptions.InvalidResponse:

        Returns:
            Union[requests.Response, dict, List] : the assets. Type: requests.Response if json_response is False, a dict
                if paginated is False, a List otherwise.
        """

        path = path or ""
        url = f"{self._base_url}/{asset_type}/{path}"
        if not url.endswith("/"):
            url = url + "/"  # server requires a suffix /

        # first request
        response = self._request(
            request_name,
            url,
            **request_kwargs,
        )

        if not json_response:
            return response

        try:
            json_resp = response.json()
            if not paginated:
                return json_resp
            # in case response is expected to be paginated, get all pages
            url_next = json_resp["next"]
            while url_next:
                resp = self._request(
                    request_name,
                    url_next,
                    **request_kwargs,
                )
                json_resp["results"].extend(resp.json()["results"])
                url_next = resp.json()["next"]
            # extract full asset list from returned structure
            return json_resp["results"]
        except ValueError as e:
            msg = f"Cannot parse response to JSON: {e}"
            raise exceptions.InvalidResponse(response, msg)

    def get(self, name, key):
        """Get asset by key."""
        return self.request(
            "get",
            name,
            path=f"{key}",
        )

    def list(
        self,
        asset_type: str,
        filters: Dict[str, Union[List[str], str, dict]] = None,
        order_by: str = None,
        ascending: bool = False,
        paginated: bool = True,
        path: str = None,
    ) -> List[Dict]:
        """List assets by filters.

        Args:
            asset_type (str): asset type. e.g. "function"
            filters (dict, optional): keys = attributes, values = value(s) for this attribute
                (depends on the attribute, cf [sdk reference](references/sdk.md)).
                e.g. {"key": ["key1", "key2"]}. "," corresponds to an "OR". Defaults to None.
                                "," corresponds to an "OR".
            order_by (str, optional): attribute name to order the results on. Defaults to None.
                e.g. "name" for an ordering on name.
            ascending (bool, optional): to reverse ordering. Defaults to False (descending order).
            paginated (bool, optional): whether the response would be paginated or not. Defaults to True.
            path (str, optional): additional route e.g. "cp_key/perf" in /compute_plan/cp_key/perf. Defaults to None.

        Returns:
            List[Dict] : a List of assets (dicts)
        """

        request_kwargs = {"params": {}}
        if filters:
            request_kwargs["params"].update(request_formatter.format_search_filters_for_remote(filters))
        if order_by:
            request_kwargs["params"]["ordering"] = request_formatter.format_search_ordering_for_remote(
                order_by, ascending
            )
        items = self.request("get", asset_type, path=path, paginated=paginated, **request_kwargs)

        return items

    def _add(self, name, **request_kwargs):
        """Add asset wrapper.

        Handles conflict error when created asset already exists.
        """
        try:
            return self.request("post", name, **request_kwargs)

        except exceptions.AlreadyExists as e:
            key = e.key
            is_many = isinstance(key, list)
            if is_many:
                logger.warning(
                    "AlreadyExists exception was received for a list of keys. "
                    "Unable to determine which key(s) already exist."
                )
                raise

            logger.warning(f"{name} already exists: key='{key}'")
            if name == schemas.Type.ComputePlan:
                # We only need to retrieve the full asset in the case of a Compute Plan.
                return self.get(name, key)
            else:
                return {"key": key}

    def add(self, name, retry_timeout=False, **request_kwargs):
        """Add asset.

        In case of timeout, block till resource is created.
        """
        try:
            return self._add(name, **request_kwargs)

        except exceptions.RequestTimeout as e:
            key = e.key
            is_many = isinstance(key, list)  # timeout on many objects is not handled
            if not retry_timeout or is_many:
                raise e

            logger.warning(f"Request timeout, blocking till {name} is created: key={key}")
            retry = utils.retry_on_exception(
                exceptions=(exceptions.RequestTimeout),
                timeout=float(retry_timeout),
            )
            # XXX as there is no guarantee that the request has been sent to the ledger
            #     (and will be processed), retry on on the add request and ignore
            #     potential conflicts
            return retry(self._add)(name, **request_kwargs)

    def _update(self, name, key, **request_kwargs):
        """Update asset wrapper."""
        return self.request("put", name, key, **request_kwargs)

    def update(self, name, key, retry_timeout=False, **request_kwargs):
        """Add asset.

        In case of timeout, block till resource is created.
        """
        try:
            return self._update(name, key, **request_kwargs)
        except exceptions.RequestTimeout as e:
            if not retry_timeout:
                raise e

            logger.warning(f"Request timeout, blocking till {name} is updated: key={key}")
            retry = utils.retry_on_exception(
                exceptions=(exceptions.RequestTimeout),
                timeout=float(retry_timeout),
            )
            # XXX as there is no guarantee that the request has been sent to the ledger
            #     (and will be processed), retry on the update request and ignore
            #     potential conflicts
            return retry(self._update)(name, key, **request_kwargs)

    def get_data(self, address, **request_kwargs):
        """Get asset data."""
        return self._request(
            "get",
            address,
            **request_kwargs,
        )



METRIC = {
    "key": "d5002e1c-d50b-d5de-5341-df8a7b7d11b6",
    "name": "Skin Lesion Classification Challenge",
    "owner": "MyOrg1MSP",
    "creation_date": "2021-08-24T13:36:07.428974249Z",
    "metadata": {},
    "permissions": {
        "process": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
        "download": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
    },
    "description": {
        "checksum": "77c7a32520f7564b03f4abac4271307cc639cd9fb78b90435328278f3f24d796",
        "storage_address": "http://testserver/metric/7b75b728-9b9b-45fe-a867-4bb1525b7b5d/description/",
    },
    "function": {
        "checksum": "52eea19fccfde2b12e30f4d16fd7d48f035e03210ad5804616f71799c4bdc0de",
        "storage_address": "http://testserver/metric/7b75b728-9b9b-45fe-a867-4bb1525b7b5d/file/",
    },
    "inputs": {
        "datasamples": {"kind": "ASSET_DATA_SAMPLE", "optional": False, "multiple": True},
        "opener": {"kind": "ASSET_DATA_MANAGER", "optional": False, "multiple": False},
        "predictions": {"kind": "ASSET_MODEL", "optional": False, "multiple": False},
    },
    "outputs": {"performance": {"kind": "ASSET_PERFORMANCE", "multiple": False}},
}

DATASET = {
    "key": "d7dc4a5b-e81f-4f3f-b94c-5e2bbaca15cd",
    "name": "c5e21113fcc64160b4dfac17892db879_test_composite_traintask_data_samples_relative_order - Dataset 0",
    "owner": "MyOrg1MSP",
    "creation_date": "2021-08-24T13:36:07.428974249Z",
    "permissions": {
        "process": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
        "download": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
    },
    "type": "Test",
    "data_sample_keys": [],
    "metadata": {"foo": "bar"},
    "opener": {
        "checksum": "9d6cdc6cc5963476f371acc3d2a6e5a68727c2cc0d8f3af6262f2bad70421887",
        "storage_address": "http://testserver/data_manager/d7dc4a5b-e81f-4f3f-b94c-5e2bbaca15cd/opener/",
    },
    "description": {
        "checksum": "0cfe73931279111b113b3c084ee3c7d4a2c84960b4bd9862ce67039fc118684e",
        "storage_address": "http://testserver/data_manager/d7dc4a5b-e81f-4f3f-b94c-5e2bbaca15cd/description/",
    },
    "logs_permission": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
}

DATA_SAMPLE = {
    "key": "7096db67-c175-4b45-9fb3-ad785eebb347",
    "owner": "MyOrg1MSP",
    "data_manager_keys": ["d7dc4a5b-e81f-4f3f-b94c-5e2bbaca15cd"],
    "creation_date": "2021-08-24T13:36:07.428974249Z",
}

FUNCTION = {
    "key": "17f98afc-2b82-4ce9-b232-1a471633d020",
    "name": "def637b111f2495bb6b4771644d2409e_test_traintask_data_samples_relative_order - Function 0",
    "owner": "MyOrg1MSP",
    "creation_date": "2021-08-24T13:36:07.393646367Z",
    "permissions": {
        "process": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
        "download": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
    },
    "metadata": {},
    "description": {
        "checksum": "756589d5971c421a388a751d533ab8ce09715c93040e9e8fff1365e831545aa2",
        "storage_address": "http://testserver/function/17f98afc-2b82-4ce9-b232-1a471633d020/description/",
    },
    "function": {
        "checksum": "aa8d43bf6e3341b0034a2e396451ab731ccca95a4c1d4f65a4fcd30f9081ec7d",
        "storage_address": "http://testserver/function/17f98afc-2b82-4ce9-b232-1a471633d020/file/",
    },
    "inputs": {
        "datasamples": {"kind": "ASSET_DATA_SAMPLE", "optional": False, "multiple": True},
        "model": {"kind": "ASSET_MODEL", "optional": True, "multiple": True},
        "opener": {"kind": "ASSET_DATA_MANAGER", "optional": False, "multiple": False},
    },
    "outputs": {"model": {"kind": "ASSET_MODEL", "multiple": False}},
}


FUNCTION_PREDICT = {
    "key": "681eedb9-db00-4480-a66f-63c86cc20280",
    "name": "def637b111f2495bb6b4771644d2409e_test_traintask_data_samples_relative_order - Function Predict",
    "owner": "MyOrg1MSP",
    "creation_date": "2021-08-24T13:36:07.393646367Z",
    "permissions": {
        "process": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
        "download": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
    },
    "metadata": {},
    "description": {
        "checksum": "756589d5971c421a388a751d533ab8ce09715c93040e9e8fff1365e831545aa2",
        "storage_address": "http://testserver/function/681eedb9-db00-4480-a66f-63c86cc20280/description/",
    },
    "function": {
        "checksum": "aa8d43bf6e3341b0034a2e396451ab731ccca95a4c1d4f65a4fcd30f9081ec7d",
        "storage_address": "http://testserver/function/681eedb9-db00-4480-a66f-63c86cc20280/file/",
    },
    "inputs": {
        "datasamples": {"kind": "ASSET_DATA_SAMPLE", "optional": False, "multiple": True},
        "model": {"kind": "ASSET_MODEL", "optional": False, "multiple": False},
        "opener": {"kind": "ASSET_DATA_MANAGER", "optional": False, "multiple": False},
        "shared": {"kind": "ASSET_MODEL", "optional": True, "multiple": False},
    },
    "outputs": {"predictions": {"kind": "ASSET_MODEL", "multiple": False}},
}


FUNCTION_METRIC = {
    "key": "6a8ada2e-740f-46f4-af0f-11376763ed72",
    "name": "def637b111f2495bb6b4771644d2409e_test_traintask_data_samples_relative_order - Function Metric",
    "owner": "MyOrg1MSP",
    "creation_date": "2021-08-24T13:36:07.393646367Z",
    "permissions": {
        "process": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
        "download": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
    },
    "metadata": {},
    "description": {
        "checksum": "756589d5971c421a388a751d533ab8ce09715c93040e9e8fff1365e831545aa2",
        "storage_address": "http://testserver/function/6a8ada2e-740f-46f4-af0f-11376763ed72/description/",
    },
    "function": {
        "checksum": "aa8d43bf6e3341b0034a2e396451ab731ccca95a4c1d4f65a4fcd30f9081ec7d",
        "storage_address": "http://testserver/function/6a8ada2e-740f-46f4-af0f-11376763ed72/file/",
    },
    "inputs": {
        "datasamples": {"kind": "ASSET_DATA_SAMPLE", "optional": False, "multiple": True},
        "opener": {"kind": "ASSET_DATA_MANAGER", "optional": False, "multiple": False},
        "predictions": {"kind": "ASSET_MODEL", "optional": False, "multiple": False},
    },
    "outputs": {"performance": {"kind": "ASSET_PERFORMANCE", "multiple": False}},
}


TRAINTASK = {
    "key": "30c283be-d385-424e-94a6-4d8538275260",
    "function": {
        "key": "17f98afc-2b82-4ce9-b232-1a471633d020",
        "name": "def637b111f2495bb6b4771644d2409e_test_traintask_data_samples_relative_order - Function 0",
        "owner": "MyOrg1MSP",
        "creation_date": "2021-08-24T13:36:07.428974249Z",
        "permissions": {
            "process": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
            "download": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
        },
        "metadata": {},
        "description": {
            "checksum": "756589d5971c421a388a751d533ab8ce09715c93040e9e8fff1365e831545aa2",
            "storage_address": "http://backend-org-1-substra-backend-server.org-1:8000/function/17f98afc-2b82-4ce9-b232-1a471633d020/description/",  # noqa: E501
        },
        "function": {
            "checksum": "aa8d43bf6e3341b0034a2e396451ab731ccca95a4c1d4f65a4fcd30f9081ec7d",
            "storage_address": "http://backend-org-1-substra-backend-server.org-1:8000/function/17f98afc-2b82-4ce9-b232-1a471633d020/file/",  # noqa: E501
        },
        "inputs": {
            "datasamples": {"kind": "ASSET_DATA_SAMPLE", "optional": False, "multiple": True},
            "model": {"kind": "ASSET_MODEL", "optional": True, "multiple": True},
            "opener": {"kind": "ASSET_DATA_MANAGER", "optional": False, "multiple": False},
        },
        "outputs": {"model": {"kind": "ASSET_MODEL", "multiple": False}},
    },
    "owner": "MyOrg1MSP",
    "creation_date": "2021-08-24T13:36:07.393646367Z",
    "compute_plan_key": "a428d7a3-7458-421d-a818-7a2448225b8d",
    "metadata": {},
    "status": "STATUS_DONE",
    "worker": "MyOrg1MSP",
    "rank": 0,
    "tag": "",
    "start_date": "2021-10-12T09:28:06.947765800Z",
    "end_date": "2021-10-12T09:30:04.705947400Z",
    "inputs": [
        {"identifier": "opener", "asset_key": "d7dc4a5b-e81f-4f3f-b94c-5e2bbaca15cd"},
        {"identifier": "datasamples", "asset_key": "3180e12c-a821-434a-ad8a-a341076c7f98"},
        {"identifier": "datasamples", "asset_key": "21bb59ca-abd4-4154-b04a-44a92556a078"},
        {"identifier": "datasamples", "asset_key": "67512646-2464-4521-84de-419b1b307d30"},
        {"identifier": "datasamples", "asset_key": "8c5e430b-cb6b-4f4e-95da-499b523d9f5b"},
    ],
    "outputs": {
        "model": {
            "permissions": {
                "process": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
                "download": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
            },
            "transient": False,
        },
    },
}

TRAINTASK_MODEL = {
    "asset": {
        "key": "6f0ee20a328044fb89e70ee5d219fa0c",
        "compute_task_key": "30c283be-d385-424e-94a6-4d8538275260",
        "address": {
            "checksum": "40483cd8b99ea7fbd3b73020997ea07547771993a6a3fa56fa2a8e9d7860529e",
            "storage_address": "http://backend-org-1-substra-backend-server.org-1:8000/model/5f0ee20a328044fb89e70ee5d219fa0b/address/",  # noqa: E501
        },
        "permissions": {
            "process": {"public": False, "authorized_ids": ["MyOrg1MSP", "MyOrg2MSP"]},
            "download": {"public": False, "authorized_ids": ["MyOrg1MSP", "MyOrg2MSP"]},
        },
        "owner": "MyOrg1MSP",
        "creation_date": "2021-08-24T13:36:07.393646367Z",
    },
    "identifier": "model",
    "kind": "ASSET_MODEL",
}

TRAINTASK_MODEL_RESPONSE = {"count": 1, "previous": None, "next": None, "results": [TRAINTASK_MODEL]}

AGGREGATETASK = {
    "key": "06207faf-1785-4fa9-4220-99a50dcfe064",
    "function": {
        "key": "7c9f9799-bf64-c100-6238-1583a9ffc535",
        "name": "Logistic regression",
        "function": {
            "checksum": "7c9f9799bf64c10002381583a9ffc535bc3f4bf14d6g0c614d3f6f868f72a9d5",
            "storage_address": "",
        },
        "description": {
            "checksum": "124a0725b746d7072282d167b53cb6aab3a31bf1946dae89135c15b0126ebec3",
            "storage_address": "",
        },
        "owner": "ab75010bacbd1a4b826dc2e9ead6f1e4e1c4feade2d62a8b708fdde48fb0edeb",
        "creation_date": "2021-08-24T13:36:07.428974249Z",
        "permissions": {
            "process": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
            "download": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
        },
        "metadata": {"foo": "bar"},
        "inputs": {"model": {"kind": "ASSET_MODEL", "optional": False, "multiple": True}},
        "outputs": {"model": {"kind": "ASSET_MODEL", "multiple": False}},
    },
    "owner": "MyOrg1MSP",
    "creation_date": "2021-08-24T13:36:07.393646367Z",
    "compute_plan_key": "",
    "metadata": {"foo": "bar"},
    "status": "STATUS_DONE",
    "worker": "MyOrg1MSP",
    "rank": None,
    "tag": "My super tag",
    "start_date": "2021-10-12T09:28:06.947765800Z",
    "end_date": "2021-10-12T09:30:04.705947400Z",
    "inputs": [],
    "outputs": {
        "model": {
            "permissions": {
                "process": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
                "download": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
            },
            "transient": False,
        },
    },
}

AGGREGATETASK_MODEL = {
    "asset": {
        "key": "5f0ee20a328044fb89e70ee5d219fa0b",
        "compute_task_key": "06207faf-1785-4fa9-4220-99a50dcfe064",
        "address": {
            "checksum": "40483cd8b99ea7fbd3b73020997ea07547771993a6a3fa56fa2a8e9d7860529e",
            "storage_address": "http://backend-org-1-substra-backend-server.org-1:8000/model/5f0ee20a328044fb89e70ee5d219fa0b/address/",  # noqa: E501
        },
        "permissions": {
            "process": {"public": False, "authorized_ids": ["MyOrg1MSP", "MyOrg2MSP"]},
            "download": {"public": False, "authorized_ids": ["MyOrg1MSP", "MyOrg2MSP"]},
        },
        "owner": "MyOrg1MSP",
        "creation_date": "2021-08-24T13:36:07.393646367Z",
    },
    "identifier": "model",
    "kind": "ASSET_MODEL",
}

AGGREGATETASK_MODEL_RESPONSE = {"count": 1, "previous": None, "next": None, "results": [AGGREGATETASK_MODEL]}

COMPOSITE_TRAINTASK = {
    "key": "aa09180a-fec6-46a5-a1a7-58c971b39217",
    "function": {
        "key": "c663b6e6-dd62-49fb-afe8-191fa7627a64",
        "name": "c5e21113fcc64160b4dfac17892db879_test_aggregate_composite_traintasks - Function 0",
        "owner": "MyOrg1MSP",
        "creation_date": "2021-08-24T13:36:07.428974249Z",
        "permissions": {
            "process": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
            "download": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
        },
        "metadata": {},
        "description": {
            "checksum": "40483cd8b99ea7fbd3b73020997ea07547771993a6a3fa56fa2a8e9d7860529e",
            "storage_address": "http://backend-org-1-substra-backend-server.org-1:8000/function/c663b6e6-dd62-49fb-afe8-191fa7627a64/description/",  # noqa: E501
        },
        "function": {
            "checksum": "51bd5fe2e7f087b203d1b4a73f3b3276b9fde96a0fff9c1f5984de96e4675d59",
            "storage_address": "http://backend-org-1-substra-backend-server.org-1:8000/function/c663b6e6-dd62-49fb-afe8-191fa7627a64/file/",  # noqa: E501
        },
        "inputs": {
            "datasamples": {"kind": "ASSET_DATA_SAMPLE", "optional": False, "multiple": True},
            "local": {"kind": "ASSET_MODEL", "optional": True, "multiple": False},
            "opener": {"kind": "ASSET_DATA_MANAGER", "optional": False, "multiple": False},
            "shared": {"kind": "ASSET_MODEL", "optional": True, "multiple": False},
        },
        "outputs": {
            "local": {"kind": "ASSET_MODEL", "multiple": False},
            "shared": {"kind": "ASSET_MODEL", "multiple": False},
        },
    },
    "owner": "MyOrg1MSP",
    "creation_date": "2021-08-24T13:36:07.393646367Z",
    "compute_plan_key": "ed74db2f-b302-4a9d-b3d3-336eb8dcc3ff",
    "metadata": {},
    "status": "STATUS_DONE",
    "worker": "MyOrg2MSP",
    "rank": 0,
    "tag": "",
    "start_date": "2021-10-12T09:28:06.947765800Z",
    "end_date": "2021-10-12T09:30:04.705947400Z",
    "inputs": [
        {"identifier": "opener", "asset_key": "1ae64423-aa99-4b8a-8660-61a56e4ca42d"},
        {"identifier": "datasamples", "asset_key": "1288d38c-dec2-4433-bc40-ec17f99f522a"},
    ],
    "outputs": {
        "shared": {
            "permissions": {
                "process": {"public": False, "authorized_ids": ["MyOrg1MSP", "MyOrg2MSP"]},
                "download": {"public": False, "authorized_ids": ["MyOrg1MSP", "MyOrg2MSP"]},
            },
            "transient": False,
        },
        "local": {
            "permissions": {
                "process": {"public": False, "authorized_ids": ["MyOrg2MSP"]},
                "download": {"public": False, "authorized_ids": ["MyOrg2MSP"]},
            },
            "transient": False,
        },
    },
}

COMPOSITE_TRAINTASK_SHARED = {
    "asset": {
        "key": "dc9d114d1ed54936b57dd7fef1c0cbf4",
        "compute_task_key": "aa09180a-fec6-46a5-a1a7-58c971b39217",
        "address": {
            "checksum": "40483cd8b99ea7fbd3b73020997ea07547771993a6a3fa56fa2a8e9d7860529e",
            "storage_address": "http://backend-org-1-substra-backend-server.org-1:8000/model/ac4a1a82ced4419b83f4bba78c024b94/address/",  # noqa: E501
        },
        "permissions": {
            "process": {"public": False, "authorized_ids": ["MyOrg1MSP", "MyOrg2MSP"]},
            "download": {"public": False, "authorized_ids": ["MyOrg1MSP", "MyOrg2MSP"]},
        },
        "owner": "MyOrg1MSP",
        "creation_date": "2021-08-24T13:36:07.393646367Z",
    },
    "identifier": "shared",
    "kind": "ASSET_MODEL",
}


COMPOSITE_TRAINTASK_SHARED_RESPONSE = {"count": 1, "previous": None, "next": None, "results": [AGGREGATETASK_MODEL]}

COMPOSITE_TRAINTASK_LOCAL = {
    "asset": {
        "key": "dc9d114d1ed54936b57dd7fef1c0cbf4",
        "compute_task_key": "aa09180a-fec6-46a5-a1a7-58c971b39217",
        "address": {
            "checksum": "40483cd8b99ea7fbd3b73020997ea07547771993a6a3fa56fa2a8e9d7860529e",
            "storage_address": "http://backend-org-1-substra-backend-server.org-1:8000/model/ac4a1a82ced4419b83f4bba78c024b94/address/",  # noqa: E501
        },
        "permissions": {
            "process": {"public": False, "authorized_ids": ["MyOrg1MSP", "MyOrg2MSP"]},
            "download": {"public": False, "authorized_ids": ["MyOrg1MSP", "MyOrg2MSP"]},
        },
        "owner": "MyOrg1MSP",
        "creation_date": "2021-08-24T13:36:07.393646367Z",
    },
    "identifier": "local",
    "kind": "ASSET_MODEL",
}


COMPOSITE_TRAINTASK_LOCAL_RESPONSE = {
    "count": 1,
    "previous": None,
    "next": None,
    "results": [COMPOSITE_TRAINTASK_LOCAL],
}

COMPOSITE_TRAINTASK_DOING = {
    "key": "aa09180a-fec6-46a5-a1a7-58c971b39217",
    "function": {
        "key": "c663b6e6-dd62-49fb-afe8-191fa7627a64",
        "name": "c5e21113fcc64160b4dfac17892db879_test_aggregate_composite_traintasks - Function 0",
        "owner": "MyOrg1MSP",
        "creation_date": "2021-08-24T13:36:07.428974249Z",
        "permissions": {
            "process": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
            "download": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
        },
        "metadata": {},
        "description": {
            "checksum": "40483cd8b99ea7fbd3b73020997ea07547771993a6a3fa56fa2a8e9d7860529e",
            "storage_address": "http://backend-org-1-substra-backend-server.org-1:8000/function/c663b6e6-dd62-49fb-afe8-191fa7627a64/description/",  # noqa: E501
        },
        "function": {
            "checksum": "51bd5fe2e7f087b203d1b4a73f3b3276b9fde96a0fff9c1f5984de96e4675d59",
            "storage_address": "http://backend-org-1-substra-backend-server.org-1:8000/function/c663b6e6-dd62-49fb-afe8-191fa7627a64/file/",  # noqa: E501
        },
        "inputs": {
            "datasamples": {"kind": "ASSET_DATA_SAMPLE", "optional": False, "multiple": True},
            "local": {"kind": "ASSET_MODEL", "optional": True, "multiple": False},
            "opener": {"kind": "ASSET_DATA_MANAGER", "optional": False, "multiple": False},
            "shared": {"kind": "ASSET_MODEL", "optional": True, "multiple": False},
        },
        "outputs": {
            "local": {"kind": "ASSET_MODEL", "multiple": False},
            "shared": {"kind": "ASSET_MODEL", "multiple": False},
        },
    },
    "owner": "MyOrg1MSP",
    "creation_date": "2021-08-24T13:36:07.393646367Z",
    "compute_plan_key": "ed74db2f-b302-4a9d-b3d3-336eb8dcc3ff",
    "metadata": {},
    "status": "STATUS_DOING",
    "worker": "MyOrg2MSP",
    "rank": 0,
    "tag": "",
    "inputs": [
        {"identifier": "opener", "asset_key": "1ae64423-aa99-4b8a-8660-61a56e4ca42d"},
        {"identifier": "datasamples", "asset_key": "1288d38c-dec2-4433-bc40-ec17f99f522a"},
    ],
    "outputs": {
        "shared": {
            "permissions": {
                "process": {"public": False, "authorized_ids": ["MyOrg1MSP", "MyOrg2MSP"]},
                "download": {"public": False, "authorized_ids": ["MyOrg1MSP", "MyOrg2MSP"]},
            },
            "transient": True,
        },
        "local": {
            "permissions": {
                "process": {"public": False, "authorized_ids": ["MyOrg2MSP"]},
                "download": {"public": False, "authorized_ids": ["MyOrg2MSP"]},
            },
            "transient": False,
        },
    },
}

PREDICTTASK = {
    "key": "afa9c7b1-21b5-4e58-a098-79de4ecede35",
    "function": {
        "key": "681eedb9-db00-4480-a66f-63c86cc20280",
        "name": "def637b111f2495bb6b4771644d2409e_test_traintask_data_samples_relative_order - Function Predict681eedb9-db00-4480-a66f-63c86cc20280",  # noqa: E501
        "owner": "MyOrg1MSP",
        "creation_date": "2021-08-24T13:36:07.428974249Z",
        "permissions": {
            "process": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
            "download": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
        },
        "metadata": {},
        "description": {
            "checksum": "756589d5971c421a388a751d533ab8ce09715c93040e9e8fff1365e831545aa2",
            "storage_address": "http://backend-org-1-substra-backend-server.org-1:8000/function/681eedb9-db00-4480-a66f-63c86cc20280/description/",  # noqa: E501
        },
        "function": {
            "checksum": "aa8d43bf6e3341b0034a2e396451ab731ccca95a4c1d4f65a4fcd30f9081ec7d",
            "storage_address": "http://backend-org-1-substra-backend-server.org-1:8000/function/681eedb9-db00-4480-a66f-63c86cc20280/file/",  # noqa: E501
        },
        "inputs": {
            "datasamples": {"kind": "ASSET_DATA_SAMPLE", "optional": False, "multiple": True},
            "model": {"kind": "ASSET_MODEL", "optional": False, "multiple": False},
            "opener": {"kind": "ASSET_DATA_MANAGER", "optional": False, "multiple": False},
            "shared": {"kind": "ASSET_MODEL", "optional": True, "multiple": False},
        },
        "outputs": {"predictions": {"kind": "ASSET_MODEL", "multiple": False}},
    },
    "owner": "MyOrg1MSP",
    "creation_date": "2021-08-24T13:36:07.393646367Z",
    "compute_plan_key": "a428d7a3-7458-421d-a818-7a2448225b8d",
    "metadata": {},
    "status": "STATUS_DONE",
    "worker": "MyOrg1MSP",
    "rank": 0,
    "tag": "",
    "start_date": "2021-10-12T09:28:06.947765800Z",
    "end_date": "2021-10-12T09:30:04.705947400Z",
    "inputs": [
        {"identifier": "opener", "asset_key": "d7dc4a5b-e81f-4f3f-b94c-5e2bbaca15cd"},
        {"identifier": "datasamples", "asset_key": "3180e12c-a821-434a-ad8a-a341076c7f98"},
        {"identifier": "datasamples", "asset_key": "21bb59ca-abd4-4154-b04a-44a92556a078"},
        {
            "identifier": "model",
            "parent_task_key": "30c283be-d385-424e-94a6-4d8538275260",
            "parent_task_output_identifier": "model",
        },
    ],
    "outputs": {
        "predictions": {
            "permissions": {
                "process": {"public": True, "authorized_ids": []},
                "download": {"public": True, "authorized_ids": []},
            },
            "transient": False,
        },
    },
}

TESTTASK = {
    "key": "afa9c7b1-21b5-4e58-a098-79de4ecede35",
    "function": {
        "key": "6a8ada2e-740f-46f4-af0f-11376763ed72",
        "name": "def637b111f2495bb6b4771644d2409e_test_traintask_data_samples_relative_order - Function Metric",
        "owner": "MyOrg1MSP",
        "creation_date": "2021-08-24T13:36:07.428974249Z",
        "permissions": {
            "process": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
            "download": {"public": True, "authorized_ids": ["MyOrg1MSP"]},
        },
        "metadata": {},
        "description": {
            "checksum": "756589d5971c421a388a751d533ab8ce09715c93040e9e8fff1365e831545aa2",
            "storage_address": "http://backend-org-1-substra-backend-server.org-1:8000/function/6a8ada2e-740f-46f4-af0f-11376763ed72/description/",  # noqa: E501
        },
        "function": {
            "checksum": "aa8d43bf6e3341b0034a2e396451ab731ccca95a4c1d4f65a4fcd30f9081ec7d",
            "storage_address": "http://backend-org-1-substra-backend-server.org-1:8000/function/6a8ada2e-740f-46f4-af0f-11376763ed72/file/",  # noqa: E501
        },
        "inputs": {
            "datasamples": {"kind": "ASSET_DATA_SAMPLE", "optional": False, "multiple": True},
            "opener": {"kind": "ASSET_DATA_MANAGER", "optional": False, "multiple": False},
            "predictions": {"kind": "ASSET_MODEL", "optional": False, "multiple": False},
        },
        "outputs": {"performance": {"kind": "ASSET_PERFORMANCE", "multiple": False}},
    },
    "owner": "MyOrg1MSP",
    "creation_date": "2021-08-24T13:36:07.393646367Z",
    "compute_plan_key": "a428d7a3-7458-421d-a818-7a2448225b8d",
    "metadata": {},
    "status": "STATUS_DONE",
    "worker": "MyOrg1MSP",
    "rank": 0,
    "tag": "",
    "start_date": "2021-10-12T09:28:06.947765800Z",
    "end_date": "2021-10-12T09:30:04.705947400Z",
    "inputs": [
        {"identifier": "opener", "asset_key": "d7dc4a5b-e81f-4f3f-b94c-5e2bbaca15cd"},
        {"identifier": "datasamples", "asset_key": "3180e12c-a821-434a-ad8a-a341076c7f98"},
        {"identifier": "datasamples", "asset_key": "21bb59ca-abd4-4154-b04a-44a92556a078"},
        {
            "identifier": "predictions",
            "parent_task_key": "05cf2e58-5862-444b-82de-cdaaead97c53",
            "parent_task_output_identifier": "predictions",
        },
    ],
    "outputs": {
        "performance": {
            "permissions": {
                "process": {"public": True, "authorized_ids": []},
                "download": {"public": True, "authorized_ids": []},
            },
            "transient": False,
        },
    },
}

TASK_LIST = [PREDICTTASK, AGGREGATETASK, COMPOSITE_TRAINTASK, TRAINTASK, TESTTASK]


COMPUTE_PLAN = {
    "key": "e983a185-5368-bd0a-0190-183af9a8e560",
    "tag": "",
    "name": "My test compute plan",
    "owner": "MyOrg1MSP",
    "creation_date": "2021-08-24T13:36:07.393646367Z",
    "start_date": "2021-08-24T13:36:37.947765800Z",
    "end_date": "2021-08-24T13:38:07.705947400Z",
    "estimated_end_date": "2021-08-24T13:38:07.705947400Z",
    "duration": "90",
    "status": "PLAN_STATUS_DONE",
    "task_count": 21,
    "waiting_count": 1,
    "todo_count": 2,
    "doing_count": 3,
    "canceled_count": 4,
    "failed_count": 5,
    "done_count": 6,
    "metadata": {"foo": "bar"},
}

COMPUTE_PLAN_PERF = {
    "results": [
        {
            "compute_task": {
                "key": "afa9c7b1-21b5-4e58-a098-79de4ecede35",
                "data_manager_key": "d7dc4a5b-e81f-4f3f-b94c-5e2bbaca15cd",
                "function_key": "17f98afc-2b82-4ce9-b232-1a471633d020",
                "rank": 1,
                "epoch": None,
                "round_idx": None,
                "data_samples": [
                    "3180e12c-a821-434a-ad8a-a341076c7f98",
                    "21bb59ca-abd4-4154-b04a-44a92556a078",
                ],
                "worker": "MyOrg1MSP",
            },
            "identifier": "performance",
            "perf": 0.673,
        },
        {
            "compute_task": {
                "key": "afa9c7b1-21b5-4e58-a098-79de4ecede35",
                "data_manager_key": "d7dc4a5b-e81f-4f3f-b94c-5e2bbaca15cd",
                "function_key": "17f98afc-2b82-4ce9-b232-1a471633d020",
                "rank": 2,
                "epoch": None,
                "round_idx": None,
                "data_samples": [
                    "3180e12c-a821-434a-ad8a-a341076c7f98",
                    "21bb59ca-abd4-4154-b04a-44a92556a078",
                ],
                "worker": "MyOrg1MSP",
            },
            "identifier": "performance",
            "perf": 0.834,
        },
        {
            "compute_task": {
                "key": "afa9c7b1-21b5-4e58-a098-79de4ecede35",
                "data_manager_key": "d7dc4a5b-e81f-4f3f-b94c-5e2bbaca15cd",
                "function_key": "17f98afc-2b82-4ce9-b232-1a471633d020",
                "rank": 3,
                "epoch": None,
                "round_idx": None,
                "data_samples": [
                    "3180e12c-a821-434a-ad8a-a341076c7f98",
                    "21bb59ca-abd4-4154-b04a-44a92556a078",
                ],
                "worker": "MyOrg1MSP",
            },
            "identifier": "performance",
            "perf": 0.956,
        },
    ],
}

MODEL = {
    "address": {
        "storage_address": "http://substra-backend.org-1.com/model/d7d88d74-d42f-43f4-affa-d0cdb99873fc/file/",
        "checksum": "b7881c547f7b80f01200fd2c41546010a15c66e2db46722dbbffa428a0bd0a26",
    },
    "compute_task_key": "b2b37c0f-de0a-4a61-9d00-58079efa9b33",
    "creation_date": "2022-03-01T11:22:35.791650Z",
    "key": "d7d88d74-d42f-43f4-affa-d0cdb99873fc",
    "owner": "MyOrg1MSP",
    "permissions": {
        "download": {"authorized_ids": ["MyOrg1MSP"], "public": True},
        "process": {"authorized_ids": ["MyOrg1MSP"], "public": True},
    },
}



import collections
import logging
import typing

from substra.sdk import exceptions
from substra.sdk import models

logger = logging.getLogger(__name__)


class InMemoryDb:
    """In memory data db."""

    def __init__(self):
        # assets stored per type and per key
        self._data = collections.defaultdict(dict)

    def add(self, asset):
        """Add an asset."""
        type_ = asset.__class__.type_
        key = getattr(asset, "key", None)
        if not key:
            key = asset.id
        if key in self._data[type_]:
            raise exceptions.KeyAlreadyExistsError(f"The asset key {key} of type {type_} has already been used.")
        self._data[type_][key] = asset
        logger.info(f"{type_} with key '{key}' has been created.")

        return asset

    def get(self, type_, key: str):
        """Return asset."""
        try:
            return self._data[type_][key]
        except KeyError:
            raise exceptions.NotFound(f"Wrong pk {key}", 404)

    def _match_asset(self, asset: models._Model, attribute: str, values: typing.Union[typing.Dict, typing.List]):
        """Checks if an asset attributes matches the given values.
        For the metadata, it checks that all the given filters returns True (AND condition)"""
        if attribute == "metadata":
            metadata_conditions = []
            for value in values:
                if value["type"] == models.MetadataFilterType.exists:
                    metadata_conditions.append(value["key"] in asset.metadata.keys())

                elif asset.metadata.get(value["key"]) is None:
                    # for is_equal and contains, if the key is not there then return False
                    metadata_conditions.append(False)

                elif value["type"] == models.MetadataFilterType.is_equal:
                    metadata_conditions.append(str(value["value"]) == str(asset.metadata[value["key"]]))

                elif value["type"] == models.MetadataFilterType.contains:
                    metadata_conditions.append(str(value["value"]) in str(asset.metadata.get(value["key"])))
                else:
                    raise NotImplementedError

            return all(metadata_conditions)

        return str(getattr(asset, attribute)) in values

    def _filter_assets(
        self, db_assets: typing.List[models._Model], filters: typing.Dict[str, typing.List[str]]
    ) -> typing.List[models._Model]:
        """Return assets matching al the given filters"""

        matching_assets = [
            asset
            for asset in db_assets
            if all(self._match_asset(asset, attribute, values) for attribute, values in filters.items())
        ]
        return matching_assets

    def list(
        self, type_: str, filters: typing.Dict[str, typing.List[str]], order_by: str = None, ascending: bool = False
    ):
        """List assets by filters.

        Args:
            asset_type (str): asset type. e.g. "function"
            filters (dict, optional): keys = attributes, values = list of values for this attribute.
                e.g. {"name": ["name1", "name2"]}. "," corresponds to an "OR". Defaults to None.
            order_by (str, optional): attribute name to order the results on. Defaults to None.
                e.g. "name" for an ordering on name.
            ascending (bool, optional): to reverse ordering. Defaults to False (descending order).

        Returns:
            List[Dict] : a List of assets (dicts)
        """
        # get all assets of this type
        assets = list(self._data[type_].values())

        if filters:
            assets = self._filter_assets(assets, filters)
        if order_by:
            assets.sort(key=lambda x: getattr(x, order_by), reverse=(not ascending))

        return assets

    def update(self, asset):
        type_ = asset.__class__.type_
        key = asset.key

        if key not in self._data[type_]:
            raise exceptions.NotFound(f"Wrong pk {key}", 404)

        self._data[type_][key] = asset
        return


db = InMemoryDb()


def get_db():
    return db



import os
import pathlib
import shutil
import sys
import tempfile
import uuid
import zipfile

import substra

from .fl_interface import FLFunctionInputs
from .fl_interface import FLFunctionOutputs
from .fl_interface import FunctionCategory
from .fl_interface import InputIdentifiers
from .fl_interface import OutputIdentifiers

DEFAULT_DATA_SAMPLE_FILENAME = "data.csv"

DEFAULT_SUBSTRATOOLS_VERSION = (
    f"latest-nvidiacuda11.8.0-base-ubuntu22.04-python{sys.version_info.major}.{sys.version_info.minor}"
)

DEFAULT_SUBSTRATOOLS_DOCKER_IMAGE = f"ghcr.io/substra/substra-tools:{DEFAULT_SUBSTRATOOLS_VERSION}"

DEFAULT_OPENER_SCRIPT = f"""
import csv
import json
import os
import substratools as tools
class TestOpener(tools.Opener):
    def get_data(self, folders):
        X, y = [], []
        for folder in folders:
            with open(os.path.join(folder, '{DEFAULT_DATA_SAMPLE_FILENAME}'), 'r') as f:
                reader = csv.reader(f)
                for row in reader:
                    X.append(int(row[0]))
                    y.append(int(row[1]))
        print(f'X: {{X}}') # a list of 1's
        print(f'y: {{y}}') # a list of 2's
        return X, y
    def fake_data(self, n_samples=None):
        if n_samples is None:
            n_samples = 1
        X = [10] * n_samples
        y = [30] * n_samples
        print(f'fake X: {{X}}')
        print(f'fake y: {{y}}')
        return X, y
"""


DEFAULT_METRIC_FUNCTION_SCRIPT = f"""
import json
import substratools as tools

@tools.register
def score(inputs, outputs, task_properties):
    y_true = inputs['{InputIdentifiers.datasamples}'][1]
    y_pred = _get_predictions(inputs['{InputIdentifiers.predictions}'])
    res = sum(y_pred) - sum(y_true)
    print(f'metrics, y_true: {{y_true}}, y_pred: {{y_pred}}, result: {{res}}')
    tools.save_performance(res, outputs['{OutputIdentifiers.performance}'])

def _get_predictions(path):
    with open(path) as f:
        return json.load(f)

if __name__ == '__main__':
    tools.execute()
"""


DEFAULT_FUNCTION_SCRIPT = f"""
import json
import substratools as tools

@tools.register
def train(inputs, outputs, task_properties):
    X = inputs['{InputIdentifiers.datasamples}'][0]
    y = inputs['{InputIdentifiers.datasamples}'][1]
    models_path = inputs.get('{InputIdentifiers.shared}', [])
    models = [_load_model(model_path) for model_path in models_path]

    print(f'Train, get X: {{X}}, y: {{y}}, models: {{models}}')

    ratio = sum(y) / sum(X)
    err = 0.1 * ratio  # Add a small error

    if len(models) == 0:
        res = dict(value=ratio + err)
    else:
        ratios = [m['value'] for m in models]
        avg = sum(ratios) / len(ratios)
        res = dict(value=avg + err)

    print(f'Train, return {{res}}')
    _save_model(res, outputs['{OutputIdentifiers.shared}'])

@tools.register
def predict(inputs, outputs, task_properties):
    X = inputs['{InputIdentifiers.datasamples}'][0]
    model = _load_model(inputs['{InputIdentifiers.shared}'])

    res = [x * model['value'] for x in X]
    print(f'Predict, get X: {{X}}, model: {{model}}, return {{res}}')

    _save_predictions(res, outputs['{OutputIdentifiers.predictions}'])

def _load_model(path):
    with open(path) as f:
        return json.load(f)

def _save_model(model, path):
    with open(path, 'w') as f:
        return json.dump(model, f)

def _save_predictions(y_pred, path):
    with open(path, 'w') as f:
        return json.dump(y_pred, f)

if __name__ == '__main__':
    tools.execute()
"""

DEFAULT_AGGREGATE_FUNCTION_SCRIPT = f"""
import json
import substratools as tools

@tools.register
def aggregate(inputs, outputs, task_properties):
    models_path = inputs.get('{InputIdentifiers.shared}', [])
    models = [_load_model(model_path) for model_path in models_path]
    print(f'Aggregate models: {{models}}')
    values = [m['value'] for m in models]
    avg = sum(values) / len(values)
    res = dict(value=avg)
    print(f'Aggregate result: {{res}}')
    _save_model(res, outputs['{OutputIdentifiers.shared}'])

@tools.register
def predict(inputs, outputs, task_properties):
    _save_predictions(0, outputs['{OutputIdentifiers.predictions}'])

def _load_model(path):
    with open(path) as f:
        return json.load(f)

def _save_model(model, path):
    with open(path, 'w') as f:
        return json.dump(model, f)

def _save_predictions(y_pred, path):
    with open(path, 'w') as f:
        return json.dump(y_pred, f)

if __name__ == '__main__':
    tools.execute()
"""

# TODO we should have a different serializer for head and trunk models
DEFAULT_COMPOSITE_FUNCTION_SCRIPT = f"""
import json
import substratools as tools

@tools.register
def train(inputs, outputs, task_properties):
    X = inputs['{InputIdentifiers.datasamples}'][0]
    y = inputs['{InputIdentifiers.datasamples}'][1]
    head_model_path = inputs.get('{InputIdentifiers.local}')
    head_model = _load_model(head_model_path) if head_model_path else None

    trunk_model_path = inputs.get('{InputIdentifiers.shared}')
    trunk_model = _load_model(trunk_model_path) if trunk_model_path else None

    print(f'Composite function train X: {{X}}, y: {{y}}, head_model: {{head_model}}, trunk_model: {{trunk_model}}')

    ratio = sum(y) / sum(X)
    err_head = 0.1 * ratio  # Add a small error
    err_trunk = 0.2 * ratio  # Add a small error

    if head_model:
        res_head = head_model['value']
    else:
        res_head = ratio

    if trunk_model:
        res_trunk = trunk_model['value']
    else:
        res_trunk = ratio

    res = dict(value= res_head + err_head), dict(value= res_trunk + err_trunk)
    print(f'Composite function train head, trunk result: {{res}}')

    _save_model(res[0], outputs['{OutputIdentifiers.local}'])
    _save_model(res[1], outputs['{OutputIdentifiers.shared}'])

@tools.register
def predict(inputs, outputs, task_properties):
    X = inputs['{InputIdentifiers.datasamples}'][0]
    head_model = _load_model(inputs['{InputIdentifiers.local}'])
    trunk_model = _load_model(inputs['{InputIdentifiers.shared}'])

    print(f'Composite function predict X: {{X}}, head_model: {{head_model}}, trunk_model: {{trunk_model}}')
    ratio_sum = head_model['value'] + trunk_model['value']
    res = [x * ratio_sum for x in X]
    print(f'Composite function predict result: {{res}}')

    _save_predictions(res, outputs['{OutputIdentifiers.predictions}'])

def _load_model(path):
    with open(path) as f:
        return json.load(f)

def _save_model(model, path):
    with open(path, 'w') as f:
        return json.dump(model, f)

def _save_predictions(y_pred, path):
    with open(path, 'w') as f:
        return json.dump(y_pred, f)

if __name__ == '__main__':
    tools.execute()
"""


DEFAULT_FUNCTION_SCRIPTS = {
    FunctionCategory.simple: DEFAULT_FUNCTION_SCRIPT,
    FunctionCategory.composite: DEFAULT_COMPOSITE_FUNCTION_SCRIPT,
    FunctionCategory.aggregate: DEFAULT_AGGREGATE_FUNCTION_SCRIPT,
    FunctionCategory.predict: DEFAULT_FUNCTION_SCRIPT,
    FunctionCategory.metric: DEFAULT_METRIC_FUNCTION_SCRIPT,
    FunctionCategory.predict_composite: DEFAULT_COMPOSITE_FUNCTION_SCRIPT,
}

DEFAULT_FUNCTION_FUNCTION_NAME = {
    FunctionCategory.simple: "train",
    FunctionCategory.composite: "train",
    FunctionCategory.aggregate: "aggregate",
    FunctionCategory.predict: "predict",
    FunctionCategory.metric: "score",
    FunctionCategory.predict_composite: "predict",
}

DEFAULT_FUNCTION_DOCKERFILE = f"""
FROM {DEFAULT_SUBSTRATOOLS_DOCKER_IMAGE}
COPY function.py .
ENTRYPOINT ["python3", "function.py", "--function-name", "{{function_name}}"]
"""

BAD_ENTRYPOINT_DOCKERFILE = f"""
FROM {DEFAULT_SUBSTRATOOLS_DOCKER_IMAGE}
COPY function.py .
ENTRYPOINT ["python3", "function.txt"]
"""

NO_ENTRYPOINT_DOCKERFILE = f"""
FROM {DEFAULT_SUBSTRATOOLS_DOCKER_IMAGE}
COPY function.py .
"""

NO_FUNCTION_NAME_DOCKERFILE = f"""
FROM {DEFAULT_SUBSTRATOOLS_DOCKER_IMAGE}
COPY function.py .
ENTRYPOINT ["python3", "function.txt", "train"]
"""

DEFAULT_PERMISSIONS = substra.sdk.schemas.Permissions(public=True, authorized_ids=[])


def zip_folder(path, destination=None):
    if not destination:
        destination = os.path.join(os.path.dirname(path), os.path.basename(path) + ".zip")
    with zipfile.ZipFile(destination, "w", zipfile.ZIP_DEFLATED) as zf:
        for root, _, files in os.walk(path):
            for f in files:
                abspath = os.path.join(root, f)
                archive_path = os.path.relpath(abspath, start=path)
                zf.write(abspath, arcname=archive_path)
    return destination


def create_archive(tmpdir, *files):
    tmpdir.mkdir()
    for path, content in files:
        with open(tmpdir / path, "w") as f:
            f.write(content)
    return zip_folder(str(tmpdir))


def random_uuid():
    return str(uuid.uuid4())


def _shorten_name(name):
    """Format asset name to ensure they match the backend requirements."""
    if len(name) < 100:
        return name
    return name[:75] + "..." + name[:20]


class Counter:
    def __init__(self):
        self._idx = -1

    def inc(self):
        self._idx += 1
        return self._idx


class AssetsFactory:
    def __init__(self, name):
        self._data_sample_counter = Counter()
        self._dataset_counter = Counter()
        self._metric_counter = Counter()
        self._function_counter = Counter()
        self._workdir = pathlib.Path(tempfile.mkdtemp(prefix="/tmp/"))
        self._uuid = name

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        shutil.rmtree(str(self._workdir), ignore_errors=True)

    def create_data_sample(self, content=None, datasets=None):
        idx = self._data_sample_counter.inc()
        tmpdir = self._workdir / f"data-{idx}"
        tmpdir.mkdir()

        content = content or "10,20"
        content = content.encode("utf-8")

        data_filepath = tmpdir / DEFAULT_DATA_SAMPLE_FILENAME
        with open(data_filepath, "wb") as f:
            f.write(content)

        datasets = datasets or []

        return substra.sdk.schemas.DataSampleSpec(
            path=str(tmpdir),
            data_manager_keys=datasets,
        )

    def create_dataset(self, permissions=None, metadata=None, py_script=None, logs_permission=None, name=None):
        idx = self._dataset_counter.inc()
        tmpdir = self._workdir / f"dataset-{idx}"
        tmpdir.mkdir()
        if name is None:
            name = _shorten_name(f"{self._uuid} - Dataset {idx}")

        description_path = tmpdir / "description.md"
        description_content = name
        with open(description_path, "w") as f:
            f.write(description_content)

        opener_path = tmpdir / "opener.py"
        with open(opener_path, "w") as f:
            f.write(py_script or DEFAULT_OPENER_SCRIPT)

        return substra.sdk.schemas.DatasetSpec(
            name=name,
            data_opener=str(opener_path),
            type="Test",
            metadata=metadata,
            description=str(description_path),
            permissions=permissions or DEFAULT_PERMISSIONS,
            logs_permission=logs_permission or DEFAULT_PERMISSIONS,
        )

    def create_function(
        self,
        category=FunctionCategory.simple,
        py_script=None,
        permissions=None,
        metadata=None,
        dockerfile_type=None,
    ):
        idx = self._function_counter.inc()
        tmpdir = self._workdir / f"function-{idx}"
        tmpdir.mkdir()
        name = _shorten_name(f"{self._uuid} - Function {idx}")

        description_path = tmpdir / "description.md"
        description_content = name
        with open(description_path, "w") as f:
            f.write(description_content)

        try:
            function_content = py_script or DEFAULT_FUNCTION_SCRIPTS[category]
        except KeyError:
            raise Exception("Invalid function category: ", category)

        function_zip = create_archive(
            tmpdir / "function",
            ("function.py", function_content),
            (
                "Dockerfile",
                (
                    BAD_ENTRYPOINT_DOCKERFILE
                    if dockerfile_type == "BAD_ENTRYPOINT"
                    else NO_ENTRYPOINT_DOCKERFILE
                    if dockerfile_type == "NO_ENTRYPOINT"
                    else NO_FUNCTION_NAME_DOCKERFILE
                    if dockerfile_type == "NO_FUNCTION_NAME"
                    else DEFAULT_FUNCTION_DOCKERFILE.format(function_name=DEFAULT_FUNCTION_FUNCTION_NAME[category])
                ),
            ),
        )

        return substra.sdk.schemas.FunctionSpec(
            name=name,
            inputs=FLFunctionInputs[category],
            outputs=FLFunctionOutputs[category],
            description=str(description_path),
            file=str(function_zip),
            permissions=permissions or DEFAULT_PERMISSIONS,
            metadata=metadata,
        )

    def create_compute_plan(self, key=None, tag="", name="Test compute plan", metadata=None):
        return substra.sdk.schemas.ComputePlanSpec(
            key=key or random_uuid(),
            traintasks=[],
            composite_traintasks=[],
            aggregatetasks=[],
            predicttasks=[],
            testtasks=[],
            tag=tag,
            name=name,
            metadata=metadata,
        )

    def add_compute_plan_tasks(self, compute_plan):
        return substra.sdk.schemas.UpdateComputePlanTasksSpec(
            traintasks=[],
            composite_traintasks=[],
            aggregatetasks=[],
            predicttasks=[],
            testtasks=[],
            key=compute_plan.key,
        )



import itertools

import pytest

from substra.sdk import exceptions
from substra.sdk import graph


@pytest.fixture
def node_graph():
    return {key: list(range(key)) for key in range(10)}


@pytest.fixture
def node_graph_linear():
    return {key: [key - 1] if key > 0 else list() for key in range(10)}


@pytest.fixture
def node_graph_straight_branch():
    """
    0       1
    |-> 2 <-|
    |   |   |
    3 <---> 4
    |
    5
    |
    6
    """
    return {
        "0": [],
        "1": [],
        "2": ["0", "1"],
        "3": ["0", "2"],
        "4": ["1", "2"],
        "5": ["3", "3"],  # duplicated link
        "6": ["5"],
    }


def test_compute_ranks(node_graph):
    visited = graph.compute_ranks(node_graph=node_graph)
    for key, rank in visited.items():
        assert key == rank


def test_compute_ranks_linear(node_graph_linear):
    visited = graph.compute_ranks(node_graph=node_graph_linear)
    for key, rank in visited.items():
        assert key == rank


def test_compute_ranks_no_correlation():
    node_graph = {key: list() for key in range(10)}
    visited = graph.compute_ranks(node_graph=node_graph)
    for _, rank in visited.items():
        assert rank == 0


def test_compute_ranks_cycle(node_graph):
    node_graph[5].append(9)
    with pytest.raises(exceptions.InvalidRequest) as e:
        graph.compute_ranks(node_graph=node_graph)

    assert "missing dependency among inModels IDs" in str(e.value)


def test_compute_ranks_closed_cycle(node_graph_linear):
    node_graph_linear[0] = [9]
    with pytest.raises(exceptions.InvalidRequest) as e:
        graph.compute_ranks(node_graph=node_graph_linear)

    assert "missing dependency among inModels IDs" in str(e.value)


def test_compute_ranks_ignore(node_graph):
    node_to_ignore = set(range(5))
    for i in range(5):
        node_graph.pop(i)
    visited = graph.compute_ranks(node_graph=node_graph, node_to_ignore=node_to_ignore)
    for key, rank in visited.items():
        assert rank == key - 5


@pytest.mark.parametrize("vertices", itertools.permutations([str(i) for i in range(6)]))
def test_compute_ranks_alpha(vertices):
    """
    0       1
    |-> 2 <-|
    |   |   |
    3 <---> 4
    |
    5
    """
    node_graph = {
        "0": [],
        "1": [],
        "2": ["0", "1"],
        "3": ["0", "2"],
        "4": ["1", "2"],
        "5": ["3"],  # duplicated link
    }
    ordered_node_graph = {v: node_graph[v] for v in vertices}
    visited = graph.compute_ranks(node_graph=ordered_node_graph)
    print(list(ordered_node_graph.keys()))
    assert visited == {
        "0": 0,
        "1": 0,
        "2": 1,
        "3": 2,
        "4": 2,
        "5": 3,
    }
